"""
Timeseries module can be used to retrieve timeseries data from the SAP iot abstract timeseries api.

Here we define some convenience wrappers for timeseries data.
"""

from __future__ import annotations

from collections.abc import Iterable
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Union, Any, Callable
import logging

import pandas as pd
from plotnine import ggplot, geom_point, aes, facet_grid, geom_line
from plotnine.themes import theme
from plotnine.scales import scale_x_datetime
from sklearn.preprocessing import StandardScaler

import sailor.assetcentral.indicators as ac_indicators
from ..utils.plot_helper import _default_plot_theme
from ..utils.timestamps import _any_to_timestamp, _calculate_nice_sub_intervals

if TYPE_CHECKING:
    from ..assetcentral.indicators import IndicatorSet
    from ..assetcentral.equipment import EquipmentSet

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


class TimeseriesDataset(object):
    """A Wrapper class to make accessing timeseries data from SAP iot more convenient."""

    def __init__(self, df: pd.DataFrame, indicator_set: IndicatorSet, equipment_set: EquipmentSet,
                 nominal_data_start: pd.Timestamp, nominal_data_end: pd.Timestamp,
                 is_normalized: bool = False):
        """
        Create a TimeseriesDataset.

        indicator_set must be an IndicatorSet containing all indicators occuring in the data columns of the
        dataframe, and equipment_set must be an EquipmentSet containing all equipments occuring in the equipment_id
        column of the data.
        """
        self._df = df.query('(timestamp >= @nominal_data_start) & (timestamp < @nominal_data_end)')
        self.is_normalized = is_normalized
        self._equipment_set = equipment_set
        self._indicator_set = indicator_set
        self.nominal_data_start = nominal_data_start
        self.nominal_data_end = nominal_data_end
        self.type = 'EQUIPMENT'  # current wrapper type is always 'equipment'

        df_equipment_ids = set(df['equipment_id'].unique())
        set_equipment_ids = set(equipment.id for equipment in equipment_set)
        if df_equipment_ids - set_equipment_ids:
            raise RuntimeError('Not all equipment ids in the data are provided in the equipment set.')
        if set_equipment_ids - df_equipment_ids:
            warnings.warn('There is no data in the dataframe for some of the equipments in the equipment set.')
            self._equipment_set = self._equipment_set.filter(id=df_equipment_ids)

        df_indicator_ids = set(df.columns) - set(self.get_index_columns())
        set_indicator_ids = set(indicator._unique_id for indicator in indicator_set)
        if df_indicator_ids - set_indicator_ids:
            raise RuntimeError('Not all indicator ids in the data are provided in the indicator set.')
        if set_indicator_ids - df_indicator_ids:
            warnings.warn('There is no data in the dataframe for some of the indicators in the indicator set.')
            self._indicator_set = self._indicator_set.filter(_unique_id=df_indicator_ids)

    def get_key_columns(self, speaking_names=False):
        """
        Return those columns of the data that identify the asset.

        Currently we only support asset type 'Equipment' so this will always return columns based on the equipment.
        In the future other types (like System) will be supported here.

        Parameters
        ----------
        speaking_names
            False, return key columns
            True, return corresponding names of key columns

        Example
        -------
        Get key columns of the indicator data set 'my_indicator_data'::

            my_indicator_data.get_key_columns()
        """
        if self.type != 'EQUIPMENT':
            raise NotImplementedError('Currently only Equipment is supported as base object for timeseries data.')

        if speaking_names:
            return ['model_name', 'equipment_name']
        else:
            return ['model_id', 'equipment_id']

    @staticmethod
    def get_time_column():
        """Return the name of the column containing the time information."""
        return 'timestamp'

    def get_feature_columns(self, speaking_names=False):
        """
        Get the names of all feature columns.

        Parameters
        ----------
        speaking_names
            False, returns feature columns of a data set
            True, returns corresponding names of feature columns

        Example
        -------
        Get Template id, Indicator group name and Indicator name of columns including indicator values in the
        data set 'my_indicator_data'::

            my_indicator_data.get_feature_columns(speaking_names=True)

        """
        if speaking_names:
            return list(self._indicator_set._unique_id_to_names().values())
        return list(self._indicator_set._unique_id_to_constituent_ids().keys())

    def get_index_columns(self, speaking_names=False):
        """Return the names of all index columns (key columns and time column)."""
        return [*self.get_key_columns(speaking_names), self.get_time_column()]

    def as_df(self, speaking_names=False):
        """
        Return the data stored within this TimeseriesDataset object as a pandas dataframe.

        By default the data is returned with opaque column headers. If speaking_names is set to true, the data is
        converted such that equipment_id and model_id are replaced by human-readable names, and the opaque
        column headers are replaced by a hierarchical index of template_id, indicator_group_name, indicator_name and
        aggregation_function.
        """
        if speaking_names:
            return self._transform(self._df)
        else:
            return self._df.set_index(self.get_index_columns())

    def _transform(self, df):
        translator = {'equipment_id': {}, 'model_id': {}}
        for equipment in self._equipment_set:
            translator['equipment_id'][equipment.id] = equipment.name
            translator['model_id'][equipment.model_id] = equipment.model_name

        static_column_mapping = {'equipment_id': 'equipment_name', 'model_id': 'model_name'}
        data = (
            df.replace(translator)
              .rename(columns=static_column_mapping)
              .set_index(self.get_index_columns(speaking_names=True))
              .rename(columns=self._indicator_set._unique_id_to_names())
        )
        data.columns = pd.MultiIndex.from_tuples(data.columns)

        return data

    def plot(self, start=None, end=None, indicator_set=None, equipment_set=None):
        """
        Plot the timeseries data stored within this wrapper.

        The plot will create different panels for each indicator_group_name and template in the data, as well as each
        indicator. Data from different equipment_set will be represented by different colors. The plotnine object
        returned by this method will be rendered in jupyter notebooks, but can also be further modified by the caller.

        Parameters
        ----------
        start
            Optional start time the timeseries data is plotted.
        end
            Optional end time the timeseries data is plotted.
        indicator_set
            Optional Indicators which are plotted.
        equipment_set
            optional equipment which indicator data is plotted.

        Returns
        -------
        plot
            Line charts of timeseries data.

        Example
        -------
        Plot all Indicators for a period from 2020-07-02 to 2020-09-01 in the data set 'my_indicator_data'::

            my_indicator_data.plot('2020-07-02','2020-09-01')
        """
        key_vars = self.get_index_columns()
        time_column = self.get_time_column()

        if indicator_set is None:
            indicator_set = self._indicator_set
        feature_vars = [indicator._unique_id for indicator in indicator_set]
        name_mapping = indicator_set._unique_id_to_names()

        if equipment_set is None:
            equipment_set = self._equipment_set
        equipment_mapping = {equipment.id: equipment.name for equipment in equipment_set}
        selected_equipment_ids = equipment_mapping.keys()

        start = _any_to_timestamp(start, default=self._df[time_column].min())
        end = _any_to_timestamp(end, default=self._df[time_column].max())

        if self._df.empty:
            raise RuntimeError('There is no data in this dataset.')

        data = self._df \
            .query(f'({time_column} >= @start) & ({time_column} <= @end)') \
            .query('equipment_id in @selected_equipment_ids') \
            .filter(items=key_vars + feature_vars)
        result_equipment_ids = set(data['equipment_id'])

        if data.empty:
            raise RuntimeError('There is no data in the dataset for the selected equipments and indicators.')

        # find equipment_set that are dropped from the plot and log them to the user
        empty_equipment_ids = set(selected_equipment_ids) - result_equipment_ids
        if empty_equipment_ids:
            warnings.warn(f'Following equipment show no data and are removed from the plot: {empty_equipment_ids}')
            selected_equipment_ids = set(selected_equipment_ids) - empty_equipment_ids
        # also indicators without data need to be removed from the plot due to unknown Y axis limits
        empty_indicators = data.columns[data.isna().all()].tolist()
        if empty_indicators:
            # Todo: speaking names in the log below? Currently using our uuid
            warnings.warn(f'Following indicators show no data and are removed from the plot: {empty_indicators}')
            feature_vars = set(feature_vars) - set(empty_indicators)

        query_timedelta = end - start
        break_interval = _calculate_nice_sub_intervals(query_timedelta, 5)  # at least 5 axis breaks
        first_break = start.floor(break_interval, ambiguous=False, nonexistent='shift_backward')
        last_break = end.ceil(break_interval, ambiguous=False, nonexistent='shift_forward')
        x_breaks = pd.date_range(first_break, last_break, freq=break_interval)

        if break_interval < pd.Timedelta('1 day'):
            date_labels = '%Y-%m-%d %H:%M:%S'
        else:
            date_labels = '%Y-%m-%d'

        facet_grid_definition = 'indicator + template + indicator_group ~ .'
        facet_assignment = dict(
            template=lambda x: x.Feature.apply(lambda row: name_mapping[row][0]),
            indicator_group=lambda x: x.Feature.apply(lambda row: name_mapping[row][1]),
            indicator=lambda x: x.Feature.apply(lambda row: name_mapping[row][2])
        )

        if isinstance(self._indicator_set, ac_indicators.AggregatedIndicatorSet):
            facet_grid_definition = 'aggregation + indicator + template + indicator_group ~ .'
            facet_assignment['aggregation'] = lambda x: x.Feature.apply(lambda row: name_mapping[row][3])

        aggregation_interval = _calculate_nice_sub_intervals(query_timedelta, 100)  # at leat 100 data points
        groupers = [*self.get_key_columns(), pd.Grouper(key=time_column, freq=aggregation_interval)]
        molten_data = (
            data.groupby(groupers)
                .agg('mean')
                .reset_index()
                .dropna(1, 'all')
                .melt(id_vars=key_vars, value_vars=feature_vars, var_name='Feature')
                .assign(**facet_assignment)
                .replace({'equipment_id': equipment_mapping})
                .rename(columns={'equipment_id': 'equipment'})
        )

        facet_row_count = len(feature_vars) + len(molten_data.groupby(['template', 'indicator_group']))

        plot = (
                ggplot(molten_data, aes(x=self.get_time_column(), y='value', color='equipment')) +
                geom_point() + geom_line() +
                facet_grid(facet_grid_definition, scales='free') +
                _default_plot_theme() +
                theme(figure_size=(10, 3 * facet_row_count)) +
                scale_x_datetime(limits=(start, end), date_labels=date_labels, breaks=x_breaks)
        )

        return plot

    def normalize(self, fitted_scaler=None, scaler=StandardScaler(copy=True, with_mean=True, with_std=True)) \
            -> tuple[TimeseriesDataset, Any]:
        """
        Normalize a data frame using scaler in normalization_factors.

        Parameters
        ----------
        fitted_scaler
            Optional fitted scaler, to be used to normalize self._df
        scaler
            Type of scaler to use for normalization. Default settings implies x -> (x-m)/s, m= mean and s=std.
            Properties are computed along the columns.

        Returns
        -------
        new_wrapper
            TimeseriesDataset with self._df updated to be the normalized dataframe.
        fitted_scaler
            Fitted scaler to be used to normalize the data.

        Example
        -------
        Get normalized values for indicators in the indicator data set 'My_indicator_data'::

            My_indicator_data.normalize()[0]
        """
        features = [column for column in self._df.columns if column not in self.get_index_columns()]
        if fitted_scaler is None and self.is_normalized:
            raise RuntimeError("There is no fitted scaler but dataset is already normalized.")
        if fitted_scaler is None:
            # normalize the data and save normalization factors to normalization_factors
            fitted_scaler = scaler.fit(self._df[features])
            LOG.debug('No scaler provided for normalization, fitting scaler to dataset: %s', fitted_scaler)

        normalized_df = self._df.copy()
        normalized_df[features] = fitted_scaler.transform(normalized_df[features])
        new_wrapper = TimeseriesDataset(normalized_df, self._indicator_set,
                                        self._equipment_set,
                                        self.nominal_data_start, self.nominal_data_end,
                                        is_normalized=True)

        return new_wrapper, fitted_scaler

    def filter(self, equipment_ids: Iterable[str] = None, start: Union[str, pd.Timestamp, datetime] = None,
               end: Union[str, pd.Timestamp, datetime] = None) -> TimeseriesDataset:
        """Return a new TimeseriesDataset extracted from an original data with filter parameters.

        Only indicator data specified in filters are returned.

        Parameters
        ----------
        equipment_ids
            Optional equipment set ids to filter timeseries data.
        start
            Optional start time of timeseries data are returned.
        end
            Optional end time until timeseries data are returned.

        Example
        -------
        Filter out indicator data for an equipment 'MyEquipmentId' from the indicator data 'My_indicator_data'::

            My_indicator_data.filter(MyEquipmentId)
        """
        if isinstance(equipment_ids, str):
            equipment_ids = [equipment_ids]

        if equipment_ids is None:
            equipment_ids = [equipment.id for equipment in self._equipment_set]
        start_time = _any_to_timestamp(start, default=self.nominal_data_start)
        end_time = _any_to_timestamp(end, default=self.nominal_data_end)
        selected_equi_set = self._equipment_set.filter(id=equipment_ids)

        selected_df = self._df.query('(equipment_id in @equipment_ids) &'
                                     '(timestamp >= @start_time) & (timestamp < @end_time)')

        if len(selected_df) == 0:
            warnings.warn('The selected filters removed all data, the resulting TimeseriesDataset is empty.')
        LOG.debug('Filtered Dataset contains %s rows.', len(selected_df))

        return TimeseriesDataset(selected_df, self._indicator_set, selected_equi_set,
                                 start_time, end_time, self.is_normalized)

    def aggregate(self,
                  aggregation_interval: Union[str, pd.Timedelta],
                  aggregation_functions: Union[Iterable[Union[str, Callable]], str, Callable] = 'mean')\
            -> TimeseriesDataset:
        """
        Aggregate the TimeseriesDataset to a fixed interval, returning a new TimeseriesDataset.

        This operation will change the unique feature IDs, as the new IDs need to encode the additional information on
        the aggregation function. Accordingly there will also be an additional column index level for the
        aggregation function on the DataFrame returned by :meth:`sailor.timeseries.wrappers.TimeseriesDataset.as_df`
        when using ``speaking_names=True``.
        Note that the resulting timeseries is not equidistant if gaps larger than the aggregation interval are
        present in the original timeseries.

        Parameters
        ----------
        aggregation_interval
            String specifying the aggregation interval, e.g. '1h' or '30min'. Follows the same rules as the ``freq``
            parameter in a ``pandas.Grouper`` object.
        aggregation_functions
            Aggregation function or iterable of aggregation functions to use.
            Each aggregation_function can be a string (e.g. 'mean', 'min' etc) or a function (e.g. np.max etc).
        """
        aggregation_interval = pd.Timedelta(aggregation_interval)
        if isinstance(aggregation_functions, str) or isinstance(aggregation_functions, Callable):
            aggregation_functions = (aggregation_functions, )

        new_indicators = []
        aggregation_definition = {}
        for indicator in self._indicator_set:
            for aggregation_function in aggregation_functions:
                new_indicator = ac_indicators.AggregatedIndicator(indicator.raw, str(aggregation_function))
                new_indicators.append(new_indicator)
                aggregation_definition[new_indicator._unique_id] = (indicator._unique_id, aggregation_function)
        new_indicator_set = ac_indicators.AggregatedIndicatorSet(new_indicators)

        grouper = [*self.get_key_columns(),
                   pd.Grouper(key=self.get_time_column(), closed='left', freq=aggregation_interval)]
        df = self._df.groupby(grouper).agg(**aggregation_definition)

        return TimeseriesDataset(df.reset_index(), new_indicator_set, self._equipment_set,
                                 self.nominal_data_start, self.nominal_data_end, self.is_normalized)

    def interpolate(self, interval: Union[str, pd.Timedelta], method='pad', **kwargs) -> TimeseriesDataset:
        """
        Interpolate the TimeseriesDataset to a fixed interval, returning a new TimeseriesDataset.

        Additional arguments for the interpolation function can be passed and are forwarded to the pandas `interpolate`
        function. The resulting TimeseriesDataset will always be equidistant with timestamps between
        `self.nominal_data_start` and `self.nominal_data_end`. However, values at these timestamps may be NA depending
        on the interpolation parameters.
        By default values will be forward-filled, with no limit to the number of interpolated points between two
        given values, and no extrapolation before the first known point. The following keyword arguments can be used to
        achieve some common behaviour:
          - method='slinear' will use linear interpolation between any two known points
          - method='index' will use a pandas interpolation method instead of the scipy-based method, which
          automatically forward-fills the last known value to the end of the time-series
          - fill_value='extrapolate' will extrapolate beyond the last known value (but not backwards before the first
          known value, only applicable to scipy-based interpolation methods.)
          - limit=`N` will limit the number of interpolated points between known points to N.
        Further details on this behaviour can be found in
        https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.interpolate.html
        """
        def _fill_group(grp):
            target_times = pd.date_range(self.nominal_data_start, self.nominal_data_end, freq=interval,
                                         closed='left').round(interval)

            new_index = pd.DatetimeIndex(target_times.union(grp.timestamp))
            with_all_timestamps = grp.set_index(self.get_time_column()).reindex(new_index).sort_index()

            if len(grp) <= kwargs.get('order', 1):
                group_identifier = [grp[key].iloc[0] for key in self.get_key_columns()]
                LOG.warning(f'Not enough datapoints for interpolation in group {group_identifier}!')
                return with_all_timestamps.loc[target_times]
            tmp = with_all_timestamps.interpolate(method=method, **kwargs).loc[target_times]
            tmp.index = tmp.index.set_names('timestamp')  # loc loses index name...
            return tmp

        interval = pd.Timedelta(interval)
        if interval > (self.nominal_data_end - self.nominal_data_start):
            raise RuntimeError('Can not interpolate to an interval larger than the data range.')

        df = (
            self._df
                .groupby(self.get_key_columns())
                .apply(_fill_group)
                .drop(columns=self.get_key_columns())
                .reset_index()
        )
        return TimeseriesDataset(df, self._indicator_set, self._equipment_set,
                                 self.nominal_data_start, self.nominal_data_end, self.is_normalized)



"""
Timeseries module can be used to retrieve timeseries data from the SAP iot abstract timeseries api.

Interfaces for retrieval are aligned with AssetCentral objects such as equipment_set and indicator_set.
Timeseries data is generally stored in a pandas dataframe, wrapped in a convenience class to make it easier
to interact with the data in AssetCentral terms (see wrappers.py for the convenience class).
"""
from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Union, BinaryIO
import logging
import warnings
import time
import json
import zipfile
import gzip
from io import BytesIO

import pandas as pd

import sailor.assetcentral.indicators as ac_indicators
from ..utils.oauth_wrapper import get_oauth_client, RequestError
from ..utils.timestamps import _any_to_timestamp, _timestamp_to_date_string
from ..utils.config import SailorConfig
from .wrappers import TimeseriesDataset
from ..utils.utils import DataNotFoundWarning

if TYPE_CHECKING:
    from ..assetcentral.indicators import IndicatorSet
    from ..assetcentral.equipment import EquipmentSet

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())

fixed_timeseries_columns = {
    '_TIME': 'timestamp',
    'modelId': 'model_id',
    'equipmentId': 'equipment_id'
}


def _start_bulk_timeseries_data_export(start_date: str, end_date: str, liot_indicator_group: str) -> str:
    LOG.debug("Triggering raw indicator data export for indicator group: %s.", liot_indicator_group)
    oauth_iot = get_oauth_client('sap_iot')
    base_url = SailorConfig.get('sap_iot', 'export_url')  # todo: figure out what to do about these urls
    request_url = f'{base_url}/v1/InitiateDataExport/{liot_indicator_group}?timerange={start_date}-{end_date}'

    resp = oauth_iot.request('POST', request_url)
    return resp['RequestId']


def _check_bulk_timeseries_export_status(export_id: str) -> bool:
    LOG.debug("Checking export status for export id: %s.", export_id)
    oauth_iot = get_oauth_client('sap_iot')
    base_url = SailorConfig.get('sap_iot', 'export_url')  # todo: figure out what to do about these urls
    request_url = f'{base_url}/v1/DataExportStatus?requestId={export_id}'

    resp = oauth_iot.request('GET', request_url)

    if resp['Status'] == 'The file is available for download.':
        return True
    elif resp['Status'] in ['Request for data download is submitted.', 'Request for data download is initiated.']:
        return False
    else:
        raise RuntimeError(resp['Status'])


def _process_one_file(ifile: BinaryIO, indicator_set: IndicatorSet, equipment_set: EquipmentSet) -> pd.DataFrame:
    # each processed file contains data for some time range (one day it seems), one indicator group and all
    # equipment_set holding any data for that group in that time period.
    # Since the user might not have requested all indicators in the group we'll filter out any results that were not
    # requested. This is complicated by the fact that it's possible that the same indicator_id is present in the
    # indicator_group through two different templates. If it is requested only through one template it needs to be
    # filtered out after parsing the csv into a pandas dataframe, and converting to a
    # columnar format (one column for each (indicator_id, indicator_group_id, template_id)).

    selected_equipment_ids = [equipment.id for equipment in equipment_set]  # noqa: F841
    df = pd.read_csv(ifile)

    df['_TIME'] = pd.to_datetime(df['_TIME'], utc=True, unit='ms', errors='coerce')
    df = df.astype({'equipmentId': str, 'modelId': str, 'indicatorGroupId': str, 'templateId': str})
    df = df.pivot(index=['_TIME', 'equipmentId', 'modelId'], columns=['indicatorGroupId', 'templateId'])

    columns_to_keep = {}
    for indicator in indicator_set:
        id_tuple = (indicator._liot_id, indicator._liot_group_id, indicator.template_id)
        if id_tuple in df.columns.to_flat_index():
            columns_to_keep[id_tuple] = indicator._unique_id
        else:
            warning = DataNotFoundWarning(f'Could not find any data for indicator {indicator}')
            warnings.warn(warning)

    df.columns = df.columns.to_flat_index()
    df = (
        df.filter(items=columns_to_keep.keys())
          .reset_index()
          .rename(columns=columns_to_keep)
          .rename(columns=fixed_timeseries_columns)
          .query('equipment_id in @selected_equipment_ids')
    )
    return df


def _get_exported_bulk_timeseries_data(export_id: str,
                                       indicator_set: IndicatorSet,
                                       equipment_set: EquipmentSet) -> pd.DataFrame:
    oauth_iot = get_oauth_client('sap_iot')
    base_url = SailorConfig.get('sap_iot', 'download_url')  # todo: figure out what to do about these urls
    request_url = f"{base_url}/v1/DownloadData('{export_id}')"

    resp = oauth_iot.request('GET', request_url, headers={'Accept': 'application/octet-stream'})

    ifile = BytesIO(resp)
    try:
        zip_content = zipfile.ZipFile(ifile)
    except zipfile.BadZipFile:
        raise RuntimeError('Downloaded file is corrupted, can not process contents.')

    frames = []
    for inner_file in zip_content.filelist:
        gzip_file = zip_content.read(inner_file)
        if not gzip_file:
            continue

        try:
            gzip_content = gzip.GzipFile(fileobj=BytesIO(gzip_file))
            frames.append(_process_one_file(gzip_content, indicator_set, equipment_set))
        except gzip.BadGzipFile:
            raise RuntimeError('Downloaded file is corrupted, can not process contents.')

    if frames:
        return pd.concat(frames)
    else:
        raise RuntimeError('Downloaded File did not have any content.')


def get_indicator_data(start_date: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                       end_date: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                       indicator_set: IndicatorSet, equipment_set: EquipmentSet) -> TimeseriesDataset:
    """
    Read indicator data for a certain time period, a set of equipments and a set of indicators.

    Parameters
    ----------
    start_date:
        Date of beginning of requested timeseries data. Time components of the date will be ignored.
    end_date:
        Date of end of requested timeseries data. Time components of the date will be ignored.
    indicator_set:
        IndicatorSet for which timeseries data is returned.
    equipment_set:
        Equipment set for which the timeseries data is read.

    Example
    -------
    Get the indicator set 'my_indicator_set' timeseries data for equipments in
    the equipment set 'my_equipment_set' for a period from '2020-07-02' to '2021-01-10'::

        get_indicator_data('2020-07-02','2021-01-10', my_indicator_set, my_equipment_set)
    """
    # some notes:
    # the bulk export api *only* works on indicator groups. No filtering for equipment_set or indicator_set.
    # so we always need to download data for the whole group. We filter on individual indicator-template combinations
    # as well as individual equipment in `_process_one_file`.
    start_date = _any_to_timestamp(start_date)
    end_date = _any_to_timestamp(end_date)

    query_groups = defaultdict(list)
    for indicator in indicator_set:
        query_groups[indicator._liot_group_id].append(indicator)

    request_ids = {}
    for indicator_group, indicator_subset in sorted(query_groups.items()):  # sorted to make query order reproducable
        formatted_start_date = _timestamp_to_date_string(start_date)
        formatted_end_date = _timestamp_to_date_string(end_date)
        try:
            request_id = _start_bulk_timeseries_data_export(formatted_start_date, formatted_end_date, indicator_group)
            request_ids[request_id] = indicator_subset
        except RequestError as e:
            try:
                error_message = json.loads(e.error_text)['message']
            except (json.JSONDecodeError, KeyError):
                raise e

            if error_message == 'Data not found for the requested date range':
                warning = DataNotFoundWarning(
                    f'No data for indicator group {indicator_group} found in the requested time interval!')
                warnings.warn(warning)
                continue
            else:
                raise e

    LOG.info('Data export triggered for %s indicator groups.', len(query_groups))

    results = pd.DataFrame(columns=['model_id', 'equipment_id', 'timestamp'])
    results.timestamp = results.timestamp.astype(pd.DatetimeTZDtype(tz='UTC'))

    print('Waiting for data export', end='')
    while True:
        print('.', end='')
        for request_id in list(request_ids):
            if _check_bulk_timeseries_export_status(request_id):
                indicator_subset = ac_indicators.IndicatorSet(request_ids.pop(request_id))
                data = _get_exported_bulk_timeseries_data(request_id, indicator_subset, equipment_set)
                results = pd.merge(results, data, on=['model_id', 'equipment_id', 'timestamp'], how='outer')

        if not request_ids:
            break

        time.sleep(5)
    print()

    wrapper = TimeseriesDataset(results, indicator_set, equipment_set, start_date, end_date)
    return wrapper



"""
System module can be used to retrieve System information from AssetCentral.

Classes are provided for individual Systems as well as groups of Systems (SystemSet).
"""
from __future__ import annotations

import itertools
# import warnings
from typing import TYPE_CHECKING, Union
from datetime import datetime
from functools import cached_property
from operator import itemgetter

import pandas as pd

from sailor import sap_iot
from .utils import _fetch_data, _add_properties, _parse_filter_parameters, AssetcentralEntity, ResultSet, \
    _ac_application_url
from .equipment import find_equipment, EquipmentSet
from .indicators import IndicatorSet
from .constants import VIEW_SYSTEMS

if TYPE_CHECKING:
    from ..sap_iot import TimeseriesDataset


@_add_properties
class System(AssetcentralEntity):
    """AssetCentral System Object."""

    # Properties (in AC terminology) are: systemId, internalId, status, systemStatusDescription, modelID, modelVersion,
    # model, shortDescription, templateID, systemProvider, systemVersion, createdOn, changedOn, source, imageURL,
    # className, classID, subclass, subclassID, systemProviderID, sourceSearchTerms, systemProviderSearchTerms,
    # publishedOn, operator, operatorID, completeness

    def __init__(self, ac_json):
        """Create a new System object."""
        self.raw = ac_json

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('systemId', None, None, None),
            'name': ('internalId', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'class_name': ('className', None, None, None),
            'model_id': ('modelID', None, None, None),
            'model_name': ('model', None, None, None),
            'status_text': ('systemStatusDescription', None, None, None),
            'template_id': ('templateID', None, None, None),
        }

    @staticmethod
    def _traverse_components(component, model_order, equipment_ids, system_ids):
        """Traverse component structure recursively, starting from `component`. Pydocstyle does not know punctuation."""
        compd = {}
        compd['key'] = (component['model'], model_order)
        compd['id'] = component['id']
        compd['name'] = component['name']
        compd['order'] = component['order']
        if component['order'] is not None:
            if component['objectType'] == 'EQU':
                equipment_ids.append(component['id'])
            else:
                system_ids.append(component['id'])
        compd['object_type'] = component['objectType']
        if 'childNodes' in component.keys():
            component['childNodes'] = sorted(component['childNodes'], key=itemgetter('model', 'order'))
            compd['child_list'] = []
            for _, comps_by_model in itertools.groupby(component['childNodes'], itemgetter('model')):
                for model_order, c in enumerate(comps_by_model):
                    compd0, equipment_ids, system_ids = System._traverse_components(c, model_order,
                                                                                    equipment_ids, system_ids)
                    compd['child_list'].append(compd0)
            compd['child_list'] = sorted(compd['child_list'], key=itemgetter('order'))
        return compd, equipment_ids, system_ids

    def _update_components(self, component):
        """Add indicators and replace id with model_id in the key."""
        if component['object_type'] == 'EQU':
            obj = self._equipments.filter(id=component['id'])[0]
            component['indicators'] = self._indicators[component['id']]
        else:
            if component['id'] == self.id:
                obj = self
            else:
                obj = self._systems.filter(id=component['id'])[0]
        component['key'] = (obj.model_id, component['key'][1])
        if 'child_list' in component.keys():
            component['child_nodes'] = {}
            for child in component['child_list']:
                self._update_components(child)
                component['child_nodes'][child['key']] = child
                del component['child_nodes'][child['key']]['key']
            del component['child_list']

    @cached_property
    def _component_tree(self):
        """Prepare component tree and cache it."""
        endpoint_url = _ac_application_url() + VIEW_SYSTEMS + f'({self.id})' + '/components'
        comps = _fetch_data(endpoint_url)[0]
        self._components, equipment_ids, system_ids = System._traverse_components(comps, 0, [], [])
        if system_ids:
            self._systems = find_systems(id=system_ids)
        else:
            self._systems = SystemSet([])
        self._indicators = {}
        if equipment_ids:
            self._equipments = find_equipment(id=equipment_ids)
            for equi in self._equipments:
                self._indicators[equi.id] = equi.find_equipment_indicators(type='Measured')
        else:
            self._equipments = EquipmentSet([])
        self._update_components(self._components)
        del self._components['key']
        return self._components

    @cached_property
    def components(self):
        """Pieces of equipment that are children of the system.

        Only top level, lower levels are ignored
        """
        # warnings.warn("deprecated: attribute will be removed as soon as multilevel system hierarchies are supported",
        #              FutureWarning)
        equipment_ids = set()
        comp_tree = self._component_tree
        for c in comp_tree['child_nodes']:
            if comp_tree['child_nodes'][c]['object_type'] == 'EQU':
                equipment_ids.add(comp_tree['child_nodes'][c]['id'])
        return self._equipments.filter(id=equipment_ids)

    @staticmethod
    def _create_selection_dictionary(comp_tree):
        """Create a selection dictionary recursively based on 'comp_tree.' Pydocstyle does not know punctuation."""
        selection = {}
        selection['object_type'] = comp_tree['object_type']
        if comp_tree['object_type'] == 'EQU':
            selection['indicators'] = comp_tree['indicators']
        if 'child_nodes' in comp_tree.keys():
            selection['child_nodes'] = []
            for child in comp_tree['child_nodes']:
                sel = System._create_selection_dictionary(comp_tree['child_nodes'][child])
                sel['key'] = child
                selection['child_nodes'].append(sel)
        return selection

    def get_indicator_data(self, start: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           end: Union[str, pd.Timestamp, datetime.timestamp, datetime.date]) -> TimeseriesDataset:
        """
        Get timeseries data for all Equipment in the System.

        This is a wrapper for :meth:`sailor.sap_iot.fetch.get_indicator_data` that limits the fetch query
        to the equipment in this System.

        Each component equipment will be returned as separate rows in the dataset,
        potentially making the dataset very sparse.


        Parameters
        ----------
        start
            Begin of time series data.
        end
            End of time series data.
        """
        all_indicators = sum((equipment.find_equipment_indicators() for equipment in self.components), IndicatorSet([]))
        return sap_iot.get_indicator_data(start, end, all_indicators, self.components)


class SystemSet(ResultSet):
    """Class representing a group of Systems."""

    _element_type = System
    _method_defaults = {
        'plot_distribution': {
            'by': 'model_name',
        },
    }

    def get_indicator_data(self, start: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           end: Union[str, pd.Timestamp, datetime.timestamp, datetime.date]) -> TimeseriesDataset:
        """
        Fetch data for a set of systems for all component equipment of each system.

        This is a wrapper for :meth:`sailor.sap_iot.fetch.get_indicator_data` that limits the fetch query
        to the equipment in this SystemSet.

        Similar to ``System.get_indicator_data`` each component will be returned as separate rows in the dataset,
        potentially making the dataset very sparse.

        Parameters
        ----------
        start
            Begin of time series data.
        end
            End of time series data.
        """
        all_equipment = sum((system.components for system in self), EquipmentSet([]))
        all_indicators = sum((equipment.find_equipment_indicators() for equipment in all_equipment), IndicatorSet([]))

        return sap_iot.get_indicator_data(start, end, all_indicators, all_equipment)

    @staticmethod
    def _fill_nones(sel_nodes, indicator_list, none_positions):
        """Fill None for indicators of missing subtrees recursively."""
        for node in sel_nodes:
            if node['object_type'] == 'EQU':
                for indicator in node['indicators']:
                    none_positions.add(len(indicator_list))
                    indicator_list.append(None)
            if 'child_nodes' in node.keys():
                SystemSet._fill_nones(node['child_nodes'], indicator_list, none_positions)

    @staticmethod
    def _map_comp_info(sel_nodes, sys_nodes, indicator_list, none_positions):
        """Map selection dictionary against component dictionary recursively."""
        for node in sel_nodes:
            if node['object_type'] == 'EQU':
                for indicator in node['indicators']:
                    if (node['key'] in sys_nodes.keys()) and (indicator in sys_nodes[node['key']]['indicators']):
                        indicator_list.append((sys_nodes[node['key']]['id'], indicator))
                    else:
                        none_positions.add(len(indicator_list))
                        indicator_list.append(None)
            if 'child_nodes' in node.keys():
                if node['key'] in sys_nodes.keys():
                    SystemSet._map_comp_info(node['child_nodes'], sys_nodes[node['key']]['child_nodes'],
                                             indicator_list, none_positions)
                else:
                    SystemSet._fill_nones(node['child_nodes'], indicator_list, none_positions)

    def _map_component_information(self, selection):
        """Map selection dictionary against component dictionary of systems in a system set."""
        system_indicators = {}
        none_positions = set()
        intersection = False
        if len(selection) == 0:
            # build selection dictionary from one of the systems
            intersection = True
            selection = System._create_selection_dictionary(self[0]._component_tree)
        for system in self:
            indicator_list = []
            SystemSet._map_comp_info(selection['child_nodes'], system._component_tree['child_nodes'],
                                     indicator_list, none_positions)
            system_indicators[system.id] = indicator_list
        if intersection:
            # keep only indicators that appear for all systems
            none_positions = list(none_positions)[::-1]
            for system in self:
                for p in none_positions:
                    del system_indicators[system.id][p]
        return system_indicators


def find_systems(*, extended_filters=(), **kwargs) -> SystemSet:
    """Fetch Systems from AssetCentral with the applied filters, return a SystemSet.

    This method supports the usual filter criteria, i.e.
    - Any named keyword arguments applied as equality filters, i.e. the name of the System property is checked
    against the value of the keyword argument. If the value of the keyword argument is an iterable (e.g. a list)
    then all objects matching any of the values in the iterable are returned.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.

    Examples
    --------
     Find all Systems with name 'MySystem'::

        find_systems(name='MySystem')

    Find all Systems which either have the name 'MySystem' or the name 'MyOtherSystem'::

        find_systems(name=['MySystem', 'MyOtherSystem'])

    If multiple named arguments are provided then *all* conditions have to match.

    Example
    -------
    Find all Systems with name 'MySystem' which also is published (status_text = 'Published')::

        find_systems(name='MySystem', status_text='Published')

    The ``extended_filters`` parameter can be used to specify filters that can not be expressed as an equality. Each
    extended_filter needs to be provided as a string, multiple filters can be passed as a list of strings. As above,
    all filter criteria need to match. Extended filters can be freely combined with named arguments. Here, too all
    filter criteria need to match for a System to be returned.

    Example
    -------
    Find all Systems with creation date higher or equal to 01.01.2020::

        find_systems(extended_filters=['created_on >= "2020-01-01"'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, System.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_SYSTEMS
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)

    return SystemSet([System(obj) for obj in object_list],
                     {'filters': kwargs, 'extended_filters': extended_filters})



"""
Workorder module can be used to retrieve Workorder information from AssetCentral.

Classes are provided for individual Workorders as well as groups of Workorders (WorkorderSet).
"""

from .utils import _fetch_data, _add_properties, _parse_filter_parameters, AssetcentralEntity, ResultSet, \
    _ac_application_url
from .constants import VIEW_WORKORDERS


@_add_properties
class Workorder(AssetcentralEntity):
    """AssetCentral Workorder Object."""

    # Properties (in AC terminology) are:
    # workOrderID, shortDescription, status, statusDescription, workOrderType, workOrderTypeDescription, priority,
    # priorityDescription, isInternal, createdBy, creationDateTime, lastChangedBy, lastChangeDateTime, longDescription,
    # startDate, endDate, progressStatus, progressStatusDescription, equipmentId, equipmentName, rootEquipmentId,
    # rootEquipmentName, locationId, coordinates, source, sourceId, operatorId, location, isSourceActive,
    # assetCoreEquipmentId, operator, internalId, modelId, plant, workCenter, basicStartDate, basicEndDate,
    # actualStartDate, actualEndDate, personResponsible

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('workOrderID', None, None, None),
            'name': ('internalId', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'long_description': ('longDescription', None, None, None),
            'equipment_id': ('equipmentId', None, None, None),
            'equipment_name': ('equipmentName', None, None, None),
            'endDate': ('end_date', None, None, None),
            'location': ('location', None, None, None),
            'plant': ('plant', None, None, None),
            'priority_description': ('priorityDescription', None, None, None),
            'start_date': ('startDate', None, None, None),
            'status_text': ('statusDescription', None, None, None),
            'type': ('workOrderType', None, None, None),
            'type_description': ('workOrderTypeDescription', None, None, None),
        }


class WorkorderSet(ResultSet):
    """Class representing a group of Workorders."""

    _element_type = Workorder
    _method_defaults = {
        'plot_distribution': {
            'by': 'equipment_name',
        },
    }


def find_workorders(*, extended_filters=(), **kwargs) -> WorkorderSet:
    """Fetch Workorders from AssetCentral with the applied filters, return a WorkorderSet.

    This method supports the usual filter criteria, i.e.
    - Any named keyword arguments applied as equality filters, i.e. the name of the Workorder property is checked
    against the value of the keyword argument. If the value of the keyword argument is an iterable (e.g. a list)
    then all objects matching any of the values in the iterable are returned.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.

    Examples
    --------
    Find all Workorders with name 'MyWorkorder'::

        find_workorders(name='MyWorkorder')

    Find all Workorders which either have the name 'MyWorkorder' or the name 'MyOtherWorkorder'::

        find_workorders(name=['MyWorkorder', 'MyOtherWorkorder'])

    Find all workorders with very high priority::

        find_workorders(priority = 20)

    If multiple named arguments are provided then *all* conditions have to match.

    Example
    -------
    Find all workorders with very high priority (20) and has progress status 'pending'(15) ::

        find_workorders(priority = 20, progressStatus = 15).


    The ``extended_filters`` parameter can be used to specify filters that can not be expressed as an equality. Each
    extended_filter needs to be provided as a string, multiple filters can be passed as a list of strings. As above,
    all filter criteria need to match. Extended filters can be freely combined with named arguments. Here, too all
    filter criteria need to match for a Workorder to be returned.

    Example
    -------
    Find all Workorders with start date higher than 2020-01-01::

        find_workorders(extended_filters=['start_date >= "2020-01-01"'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, Workorder.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_WORKORDERS
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return WorkorderSet([Workorder(obj) for obj in object_list],
                        {'filters': kwargs, 'extended_filters': extended_filters})



"""
Location module can be used to retrieve Location information from AssetCentral.

Classes are provided for individual Locations as well as groups of Locations (LocationSet).
"""


from .utils import _fetch_data, _add_properties, _parse_filter_parameters, AssetcentralEntity, ResultSet, \
    _ac_application_url
from .constants import VIEW_LOCATIONS


@_add_properties
class Location(AssetcentralEntity):
    """AssetCentral Location Object."""

    # Properties (in AC terminology) are:
    # locationId, name, status, version, hasInRevision, shortDescription, location, completeness, createdOn,
    # changedOn, publishedOn, source, imageURL, locationStatus, locationTypeDescription, locationType

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('locationId', None, None, None),
            'name': ('name', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'type': ('locationType', None, None, None),
            'type_description': ('locationTypeDescription', None, None, None),
        }


class LocationSet(ResultSet):
    """Class representing a group of Locations."""

    _element_type = Location
    _method_defaults = {
        'plot_distribution': {
            'by': 'type',
        },
    }


def find_locations(*, extended_filters=(), **kwargs) -> LocationSet:
    """Fetch Locations from AssetCentral with the applied filters, return a LocationSet.

    This method supports the usual filter criteria, i.e.
    - Any named keyword arguments applied as equality filters, i.e. the name of the Location property is checked
    against the value of the keyword argument. If the value of the keyword argument is an iterable (e.g. a list)
    then all objects matching any of the values in the iterable are returned.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.


    Examples
    --------
    Find all Location with name 'MyLocation'::

      find_locations(name='MyLocation')


    Find all Locations which either have the name 'MyLocation' or the name 'MyOtherLocation'::

        find_locations(name=['MyLocation', 'MyOtherLocation'])


    If multiple named arguments are provided then *all* conditions have to match.

    Example
    -------
    Find all Locations with name 'MyLocation' which also have the location type description 'Functional Location'::

        find_locations(name='MyLocation', type_description='Functional Location')


    The ``extended_filters`` parameter can be used to specify filters that can not be expressed as an equality. Each
    extended_filter needs to be provided as a string, multiple filters can be passed as a list of strings. As above,
    all filter criteria need to match. Extended filters can be freely combined with named arguments. Here, too all
    filter criteria need to match for a Location to be returned.

    Example
    -------
    Find all Locations with a short description not matching to 'Location 1'::

        find_locations(extended_filters=['short_description != "Location 1"'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, Location.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_LOCATIONS

    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return LocationSet([Location(obj) for obj in object_list],
                       {'filters': kwargs, 'extended_filters': extended_filters})



"""Utility functions for timestamp parsing."""

import datetime
import warnings

import pandas as pd


def _odata_to_timestamp_parser(name):
    return lambda self: pd.Timestamp(float(self.raw[name][6:-2])/1000, tz='UTC') if self.raw[name] else None


def _string_to_timestamp_parser(name, unit=None):
    return lambda self: pd.Timestamp(self.raw[name], unit=unit, tz='UTC') if self.raw[name] else None


def _any_to_timestamp(value, default: pd.Timestamp = None):
    """Try to parse a timestamp provided in a variety of formats into a uniform representation as pd.Timestamp."""
    if value is None:
        return default

    if isinstance(value, str):
        timestamp = pd.Timestamp(value)
    elif isinstance(value, datetime.datetime):
        timestamp = pd.Timestamp(value)
    elif isinstance(value, datetime.date):
        timestamp = pd.Timestamp(value)
    elif isinstance(value, pd.Timestamp):
        timestamp = value
    else:
        raise RuntimeError('Can only parse ISO 8601 strings, pandas timestamps or python native timestamps.')

    if timestamp.tzinfo:
        timestamp = timestamp.tz_convert('UTC')
    else:
        warnings.warn('Trying to parse non-timezone-aware timestamp, assuming UTC.', stacklevel=2)
        timestamp = timestamp.tz_localize('UTC', ambiguous='NaT', nonexistent='NaT')

    return timestamp


def _timestamp_to_isoformat(timestamp: pd.Timestamp):
    """Return an iso-format string of a timestamp after conversion to UTC and without the timezone information."""
    if timestamp.tzinfo:
        timestamp = timestamp.tz_convert('UTC')
    return timestamp.tz_localize(None).isoformat()


def _timestamp_to_date_string(timestamp: pd.Timestamp):
    """Return a date-string (YYYY-MM-DD) from a pandas Timestamp."""
    if timestamp.tzinfo:
        timestamp = timestamp.tz_convert('UTC')
    timestamp = timestamp.tz_localize(None)
    if timestamp.date() != timestamp:
        warnings.warn('Casting timestamp to date, this operation will loose time-of-day information.')
    return str(timestamp.date())


def _calculate_nice_sub_intervals(interval, n_breaks):
    # helper to calculate 'nice' intervals breaking a given interval into *at least* n_breaks sub-intervals.
    # it would be nice to have weeks/months rather than the 7d/30d but that seems tricky --
    # see eg https://github.com/pandas-dev/pandas/issues/15303
    good_intervals = ['1s', '5s', '15s', '1min', '5min', '15min', '1h', '4h', '12h', '1d', '3d', '7d', '30d']
    good_intervals = [pd.Timedelta(x) for x in good_intervals]

    target_break_interval = interval / n_breaks
    target_break_interval = max(target_break_interval, min(good_intervals))
    freq = max(filter(lambda x: x <= target_break_interval, good_intervals))
    return freq



"""Please store all the constants here."""

VIEW_MODELS = '/services/api/v1/models'
VIEW_EQUIPMENT = '/services/api/v1/equipment'
VIEW_NOTIFICATIONS = '/services/api/v1/notification'
VIEW_OBJECTS = '/services/api/v1/objects/'
VIEW_MODEL_INDICATORS = '/services/api/v1/model'
VIEW_FAILUREMODES = '/services/api/v1/failuremodes'
VIEW_LOCATIONS = '/services/api/v1/location'
VIEW_WORKORDERS = '/services/api/v1/workorders'
VIEW_SYSTEMS = '/services/api/v1/systems'



from .equipment import find_equipment
from .model import find_models
from .failure_mode import find_failure_modes
from .location import find_locations
from .notification import find_notifications
from .system import find_systems
from .workorder import find_workorders

__all__ = ['find_equipment', 'find_models', 'find_failure_modes', 'find_locations', 'find_notifications',
           'find_systems', 'find_workorders']



"""
Retrieve Model information from AssetCentral.

Classes are provided for individual Models as well as groups of Models (ModelSet).
Models can be of type Equipment, System or FunctionalLocation, but the type is not part of the AC response currently.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from .constants import VIEW_MODEL_INDICATORS, VIEW_MODELS
from .indicators import Indicator, IndicatorSet
from .equipment import find_equipment
from .utils import _fetch_data, _add_properties, _parse_filter_parameters, \
    _apply_filters_post_request, _ac_application_url, AssetcentralEntity, ResultSet
from ..utils.timestamps import _string_to_timestamp_parser

if TYPE_CHECKING:
    from .equipment import EquipmentSet


@_add_properties
class Model(AssetcentralEntity):
    """AssetCentral Model object."""

    # Properties (in AC terminology) are:  # as returned by 'models'-api, not model-details
    # modelId, name, internalId, status, version, hasInRevision, templateId, modelTemplate, subclass,
    # generation, manufacturer, shortDescription, longDescription, completeness, createdOn, changedOn,
    # imageURL, publishedOn, source, equipmentTracking, serviceExpirationDate, modelExpirationDate,
    # releaseDate, isManufacturerValid, image, isClientValid, consume, primaryExternalId, modelSearchTerms,
    # sourceSearchTerms, manufacturerSearchTerms, class

    # Additional properties returned in model-details (again ac terminology, some have sub-structure):
    # organizationID, calibrationDate, orderStopDate, noSparePartsDate, globalId, keywords, safetyRiskCode,
    # description, descriptions[], gtin, brand, isFirmwareCompatible, templates[], classId, subclassId, adminData{},
    # sectionCompleteness{}, modelType, countryCode, referenceId, metadata, templatesDetails[]

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('modelId', None, None, None),
            'name': ('name', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'long_description': ('longDescription', None, None, None),
            'generation': ('generation', None, None, None),
            'manufacturer': ('manufacturer', None, None, None),
            'model_expiration_date': ('modelExpirationDate',
                                      _string_to_timestamp_parser('modelExpirationDate', 'ms'), None, None),
            'model_template_id': ('modelTemplate', None, None, None),
            'service_expiration_date': ('serviceExpirationDate',
                                        _string_to_timestamp_parser('serviceExpirationDate', 'ms'), None, None),
            'template_id': ('templateId', None, None, None),
        }

    def find_equipment(self, *, extended_filters=(), **kwargs) -> EquipmentSet:
        """
        Get a list of equipment derived from this Model.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all Equipment for Model 'myModelName'. Return an EquipmentSet::

           model = find_models(name='myModelName')
           model[0].find_equipment()

        The resulting Equipments can further be filter based on their properties (name, location etc).
        """
        kwargs['model_id'] = self.id
        return find_equipment(extended_filters=extended_filters, **kwargs)

    def find_model_indicators(self, *, extended_filters=(), **kwargs) -> IndicatorSet:
        """Return all Indicators assigned to the Model.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all indicators for Model 'myModelName'::

            models = find_models(name='myModelName')
            models[0].find_model_indicators()
        """
        endpoint_url = _ac_application_url() + VIEW_MODEL_INDICATORS + f'({self.id})' + '/indicatorvalues'

        # AC-BUG: this api doesn't support filters (thank you AC) so we have to fetch all of them and then filter below
        object_list = _fetch_data(endpoint_url)
        filtered_objects = _apply_filters_post_request(object_list, kwargs, extended_filters,
                                                       Indicator.get_property_mapping())

        return IndicatorSet([Indicator(obj) for obj in filtered_objects])


class ModelSet(ResultSet):
    """Class representing a group of Models."""

    _element_type = Model
    _method_defaults = {
        'plot_distribution': {
            'by': 'model_template_id',
        },
    }


def find_models(*, extended_filters=(), **kwargs) -> ModelSet:
    """Fetch Models from AssetCentral with the applied filters, return an ModelSet.

    This method supports the usual filter criteria, i.e.
    Any named keyword arguments applied as equality filters, i.e. the name of the Model property is checked
    against the value of the keyword argument. If the value of the keyword argument is an iterable (e.g. a list)
    then all objects matching any of the values in the iterable are returned.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.

    Examples
    --------
    Find all Models with name 'MyModel'::

        find_models(name='MyModel')

    Find all Models which either have the name 'MyModel' or the name 'MyOtherModel'::

        find_models(name=['MyModel', 'MyOtherModel'])


    If multiple named arguments are provided then *all* conditions have to match.

    Example
    -------
    Find all Models with name 'MyModel' which also have the short description 'Description'::

        find_models(name='MyModel', short_description='Description')

    The ``extended_filters`` parameter can be used to specify filters that can not be expressed as an equality. Each
    extended_filter needs to be provided as a string, multiple filters can be passed as a list of strings. As above,
    all filter criteria need to match. Extended filters can be freely combined with named arguments. Here, too, all
    filter criteria need to match for an Model to be returned.

    Example
    -------
    Find all Models with an expiration date before January 1, 2018::

        find_models(extended_filters=['model_expiration_date < "2018-01-01"'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, Model.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_MODELS
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return ModelSet([Model(obj) for obj in object_list],
                    {'filters': kwargs, 'extended_filters': extended_filters})



"""Improvements to OAuth and OData clients."""
import logging
import json
import warnings
from datetime import datetime, timezone
import time

from furl import furl
from rauth import OAuth2Service
import jwt

from ..config import SailorConfig
from .scope_config import SCOPE_CONFIG

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


class OAuth2Client():
    """Provide session management for OAuth2 enhanced requests :class:`~requests.sessions.Session`'s.

    Manages a single session that can be used for making requests against endpoints which accept tokens issued by
    the auth server used by this client.

    Single entrypoint should be the convenience :meth:`request` method which acts as a convenience method for making
    OAuth2 enhanced HTTP requests with an instance of this class.
    """

    def __init__(self, name, scope_config=None):
        """
        Create a OAuth2Client.

        :param name: name of the OAuth2Client. Must be a service name found in the SailorConfig.
        :param scope_config: restrict access for this instance to certain scopes listed in scope_config
        """
        self.name = name
        self.client_id = SailorConfig.get(name, 'client_id')
        self.client_secret = SailorConfig.get(name, 'client_secret')
        self.access_token_url = SailorConfig.get(name, 'access_token_url')
        self.subdomain = SailorConfig.get(name, 'subdomain')
        if 'https://' in self.access_token_url:
            self.oauth_url = 'https://' + self.subdomain + '.' + self.access_token_url[len('https://'):]
        else:
            self.oauth_url = 'https://' + self.subdomain + '.' + self.access_token_url

        scope_config = SCOPE_CONFIG if scope_config is None else scope_config
        self.configured_scopes = scope_config.get(self.name, [])
        self.resolved_scopes = []
        self._active_session = None

    def request(self, method, url, **req_kwargs):
        """Make a request using this convenience wrapper.

        The interface is the same as the :meth:`~requests.sessions.Session.requests` method provides.
        Adds OData param url formatting for GET requests.
        If headers are not set, requests for JSON content by default.

        Will use the currently attached session with this client or create a new one. If scopes are configured with
        this client, will try to resolve the scopes with the auth server first before making the request.

        Returns
        -------
        A dict if the response contains JSON. Otherwise returns the response content as bytes.

        Raises
        ------
        RequestError
            When the reponse is retrieved but response code is 400 or higher.
        """
        if method == 'GET':
            parameters = req_kwargs.pop('params', {})
            url_obj = furl(url)
            url_obj.args = {**url_obj.args, **parameters}
            url = url_obj.tostr(query_quote_plus=False)

        req_kwargs.setdefault('headers', {'Accept': 'application/json'})

        if self.configured_scopes and not self.resolved_scopes:
            try:
                self._resolve_configured_scopes()
            except Exception as exc:
                warnings.warn('Could not resolve the configured scopes. Trying to continue without scopes...')
                LOG.debug(exc, exc_info=True)

        scope = ' '.join(self.resolved_scopes) if self.resolved_scopes else None
        session = self._get_session(scope=scope)

        LOG.debug('Calling %s', url)
        response = session.request(method, url, **req_kwargs)
        if response.ok:
            if response.headers['Content-Type'] == 'application/json':
                return response.json()
            else:
                return response.content
        else:
            msg = f'Request failed. Response {response.status_code} ({response.reason}): {response.text}'
            raise RequestError(msg, response.status_code, response.reason, response.text)

    def _get_session(self, scope=None):
        """
        Return the current active session or create a new one.

        If a session exists, check if the session is valid and return that session.
        Otherwise create a new session.
        """
        if self._active_session:
            use_active_session = True
            decoded_token = jwt.decode(self._active_session.access_token_response.json()['access_token'],
                                       options={'verify_signature': False})
            expiration_time = decoded_token['exp']
            if expiration_time - time.time() < 5*60:
                LOG.debug('OAuth session expires at %s', datetime.fromtimestamp(expiration_time, tz=timezone.utc))
                use_active_session = False
            elif scope is not None:
                if sorted(decoded_token['scope']) != sorted(scope.split(' ')):
                    use_active_session = False
                    LOG.debug('Scopes are not identical.')

            if use_active_session:
                return self._active_session
            else:
                try:
                    self._active_session.close()
                except Exception:
                    LOG.exception('Could not close OAuth2Session.')

        LOG.debug('Creating new OAuth session for "%s"', self.name)
        params = {'grant_type': 'client_credentials', 'scope': scope}
        service = OAuth2Service(name=self.name, client_id=self.client_id, client_secret=self.client_secret,
                                access_token_url=self.oauth_url)
        self._active_session = service.get_auth_session('POST', data=params, decoder=json.loads)

        # the get_auth_session method of rauth does not check whether the response was 200 or not
        # and therefore does not log a proper error message
        if self._active_session.access_token_response.ok:
            return self._active_session
        else:
            self._active_session = None
            raise RuntimeError('get_auth_session call did not receive a successful token response.')

    def _resolve_configured_scopes(self):
        """
        Resolve oauth scopes based on initialized scope config.

        Involves a remote call to fetch the auth token.
        Resolved scopes will be stored in `self.resolved_scopes`.
        If no scopes are configured then calling this method has no effect.

        Returns: None
        """
        if not self.configured_scopes:
            return

        encoded_token = self._get_session().access_token_response.json()['access_token']
        decoded_token = jwt.decode(encoded_token, options={'verify_signature': False})
        all_scopes = decoded_token['scope']

        resolved_scopes = []
        missing_corresponding_scopes = []

        for short_scope in self.configured_scopes:
            matches = [scope for scope in all_scopes if scope.endswith(short_scope)]

            if len(matches) >= 1:
                # assumption:
                # when there are multiple prefixed scopes for the given
                # unprefixed scope we take the first match only
                resolved_scopes.append(matches[0])
            else:
                missing_corresponding_scopes.append(repr(short_scope))

        # assumption:
        # we consider the scope configuration invalid when at least one
        # corresponding prefixed scope from auth token is absent
        if missing_corresponding_scopes:
            warnings.warn('Could not resolve all scopes. Scope configuration considered invalid. ' +
                          f'Continuing without resolved scopes. Missing scopes: {missing_corresponding_scopes}.')
            self.resolved_scopes = []
        else:
            self.resolved_scopes = resolved_scopes


class RequestError(Exception):
    """Exception object with additional information about the status returned by a REST request."""

    def __init__(self, msg, status_code, reason, error_text):
        super().__init__(msg)
        self.status_code = status_code
        self.reason = reason
        self.error_text = error_text



"""Provides configuration management for Sailor."""
from collections import namedtuple
from contextlib import contextmanager
import os
import logging
import json
import sys
import warnings

import yaml

from .utils import DataNotFoundWarning

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())

CONFIG_PROPERTIES = ('asset_central', 'sap_iot')


@contextmanager
def try_log(exception, msg):
    """Run code in try-except clause, logs message when exception is called and re-raises exception."""
    try:
        yield
    except exception as exc:
        if hasattr(msg, "__call__"):
            msg = msg(exc)
        LOG.error(msg)
        raise


class SailorConfig(namedtuple('SailorConfig', CONFIG_PROPERTIES)):
    """Stores the config of Sailor."""

    config = None

    @staticmethod
    def get(*keys):
        """Return a value from the config.

        Supports deep lookup if values are dicts.

        Example
        -------
        >>> SailorConfig.get('asset_central', 'client_id')
        """
        if SailorConfig.config is None:
            SailorConfig.load()
        res = getattr(SailorConfig.config, keys[0])
        for k in keys[1:]:
            res = res[k]
        return res

    @staticmethod
    def load():
        """Load config of Sailor from environment or YAML file.

        Tries environment first, then YAML file.
        If there is an error during one attempt the load will fail (no further methods will be tried).

        Returns
        -------
        SailorConfig
            The loaded config. If load() was called before, returns cached config.
        """
        if SailorConfig.config is not None:
            LOG.debug('config already loaded. returning cached config')
            return SailorConfig.config

        SailorConfig.config = SailorConfig._load()
        _configure_sailor()
        return SailorConfig.config

    @staticmethod
    def _load():
        """Load config from ENV or YAML file."""
        if os.getenv('SAILOR_CONFIG_JSON') is None:
            LOG.debug('SAILOR_CONFIG_JSON not found in env. skipping config load from env.')
        else:
            LOG.debug('found SAILOR_CONFIG_JSON in env. trying to load config from env.')
            with try_log(Exception, 'Error while loading the configuration from SAILOR_CONFIG_JSON.'):
                SailorConfig.config = SailorConfig.from_env()
            LOG.info('Successfully loaded config from environment.')
            return SailorConfig.config

        yaml_paths = [os.getenv('SAILOR_CONFIG_PATH', os.path.join(os.path.abspath(os.curdir), 'config.yml')),
                      os.path.join(os.path.abspath(os.path.dirname(sys.argv[0])), 'config.yml')]
        yaml_paths = [p for p in yaml_paths if os.path.exists(p)]
        try:
            yaml_config_path = yaml_paths[0]
        except IndexError:
            LOG.debug('no config YAML file found. skipping YAML file load.')
        else:
            LOG.debug('found config file at %s. trying YAML load', yaml_config_path)
            with try_log(Exception, f'Error while loading the configuration from YAML file at {yaml_config_path}.'):
                SailorConfig.config = SailorConfig.from_yaml(yaml_config_path)
            LOG.info('Successfully loaded config from YAML file.')
            return SailorConfig.config

        raise RuntimeError('No methods left for loading the config.')

    @classmethod
    def from_env(cls):
        """Load config from environment.

        Uses ``SAILOR_CONFIG_JSON`` in environment. Value needs to be JSON encoded.
        """
        config_dict = json.loads(os.environ['SAILOR_CONFIG_JSON'])
        with try_log(TypeError, lambda e: 'Missing configuration parameter(s): %s' % str(e)[str(e).find(':')+2:]):
            return cls(**config_dict)

    @classmethod
    def from_yaml(cls, path):
        """Load config from YAML file."""
        with open(path, 'r') as f:
            config_dict = yaml.safe_load(f)
        with try_log(TypeError, lambda e: 'Missing configuration parameter(s): %s' % str(e)[str(e).find(':')+2:]):
            return cls(**config_dict)


def _configure_sailor():
    warnings.filterwarnings("always", category=DataNotFoundWarning,
                            append=True)  # simpler for the user to override this setting



"""Other utility functions that don't fit into any of the specific modules."""


# this warning concerns the interactive use case
class DataNotFoundWarning(Warning):
    """Use this warning to indicate that a query-like function returned an empty result."""

    def __init__(self, message=None):
        message = "No data found for given parameters." if message is None else message
        super().__init__(message)



"""Utility functions useful for all parts of sailor."""

from io import BytesIO

from matplotlib import pyplot as plt
import plotnine as p9


def _p9_to_svg(plotnine_plot):
    """Convert a plotnine plot to an svg string (for inclusion in html output)."""
    matplotlib_plot = plotnine_plot.draw()
    buffer = BytesIO()
    matplotlib_plot.savefig(buffer, format='svg', bbox_inches='tight')
    plt.close(matplotlib_plot)
    return buffer.getvalue().decode()


def _default_plot_theme():
    """Provide a default plot theme for out plots."""
    return p9.theme(axis_text_x=p9.element_text(rotation=45, ha='right'),
                    axis_title_x=p9.element_text(margin={'t': 20}),
                    axis_title_y=p9.element_text(margin={'r': 20}),
                    strip_text_y=p9.element_text(size=7, margin={'l': 5, 'r': 7}),
                    strip_margin=0.05,
                    panel_spacing_y=0.2,
                    )



"""
Indicators module can be used to retrieve Indicator information from AssetCentral.

Classes are provided for individual Indicators as well as groups of Indicators (IndicatorSet).
Note that the indicators here represent 'materialized' indicators, i.e. indicators attached to an equipment.
Hence they contain information on indicator_group and template used to attach it to the equipment. Currently there
is no support for unrealized 'Indicator Templates'.
"""
import hashlib
from functools import cached_property

from .utils import _add_properties, AssetcentralEntity, ResultSet


@_add_properties
class Indicator(AssetcentralEntity):
    """AssetCentral Indicator Object."""

    # Properties (in AC terminology) are: categoryID, propertyId, indicatorGroupName, indicatorName,
    # indicatorGroupDesc, objectId, indicatorSource, aggUpdatedTimestamp, indicatorCategory, colorCode,
    # thresholdDescription, convertedAggregatedValue, trend, dataType, indicatorType, convertedUOMDesc,
    # UOM, convertedUOM, indicatorDesc, indicatorColorCode, isFavorite, UOMDescription, Dimension, DimensionDesc,
    # uomdescription, dimension, dimensionDesc, pstid, uom
    # TODO update field list - I think e.g. template id is also part of hte API response

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        # TODO: There is still some weird stuff here, e.g. UOM vs. uom or convertedXXX
        return {
            'id': ('propertyId', None, None, None),
            'name': ('indicatorName', None, None, None),
            'description': ('indicatorDesc', None, None, None),
            'dimension_description': ('DimensionDescription', None, None, None),
            'dimension_id': ('Dimension', None, None, None),
            'indicator_group_description': ('indicatorGroupDesc', None, None, None),
            'indicator_group_id': ('pstid', None, None, None),
            'indicator_group_name': ('indicatorGroupName', None, None, None),
            'model_id': ('objectId', None, None, None),
            'template_id': ('categoryID', None, None, None),
            'type': ('indicatorType', None, None, None),
            'uom': ('UOM', None, None, None),
            'uom_description': ('UOMDescription', None, None, None),
            '_liot_id': ('propertyId', lambda self: 'I_' + self.raw.get('propertyId', None), None, None),
            '_liot_group_id': ('pstid', lambda self: 'IG_' + self.raw.get('pstid', None), None, None),
        }

    @cached_property
    def _unique_id(self):
        m = hashlib.sha256()
        unique_string = self.id + self.indicator_group_id + self.template_id
        m.update(unique_string.encode())
        return m.hexdigest()

    def __eq__(self, other):
        """Determine whether two (materialized) indicator instances are equal."""
        return isinstance(other, self.__class__) and other._unique_id == self._unique_id

    def __hash__(self):
        """Hash of an indicator object is the hash of it's unique id."""
        return self._unique_id.__hash__()


class AggregatedIndicator(Indicator):
    """An extension of the AssetCentral Indicator object that additionally holds aggregation information."""

    def __init__(self, ac_json, aggregation_function):
        super(AggregatedIndicator, self).__init__(ac_json)
        self.aggregation_function = aggregation_function

    @cached_property
    def _unique_id(self):
        m = hashlib.sha256()
        unique_string = self.id + self.indicator_group_id + self.template_id + self.aggregation_function
        m.update(unique_string.encode())
        return m.hexdigest()


class IndicatorSet(ResultSet):
    """Class representing a group of Indicators."""

    _element_type = Indicator
    _method_defaults = {
        'plot_distribution': {
            'by': 'indicator_group_name',
        },
    }

    def _unique_id_to_names(self):
        """Get details on an opaque column_id in terms of AssetCentral names."""
        mapping = {}
        for indicator in self:
            mapping[indicator._unique_id] = (
                indicator.template_id,  # apparently fetching the template name would need a remote call
                indicator.indicator_group_name,
                indicator.name,
            )
        return mapping

    def _unique_id_to_constituent_ids(self):
        """Get details on an opaque column_id in terms of AssetCentral IDs."""
        mapping = {}
        for indicator in self:
            mapping[indicator._unique_id] = (
                indicator.template_id,
                indicator.indicator_group_id,
                indicator.id,
            )
        return mapping


class AggregatedIndicatorSet(IndicatorSet):
    """Class representing a group of AggregatedIndicators."""

    def _unique_id_to_names(self):
        """Get details on an opaque column_id in terms of AssetCentral names and aggregation_function."""
        mapping = {}
        for indicator in self:
            mapping[indicator._unique_id] = (
                indicator.template_id,  # apparently fetching the template name would need a remote call
                indicator.indicator_group_name,
                indicator.name,
                indicator.aggregation_function,
            )
        return mapping

    def _unique_id_to_constituent_ids(self):
        """Get details on an opaque column_id in terms of AssetCentral IDs and aggregation_function."""
        mapping = {}
        for indicator in self:
            mapping[indicator._unique_id] = (
                indicator.template_id,
                indicator.indicator_group_id,
                indicator.id,
                indicator.aggregation_function,
            )
        return mapping

# while there is a generic '/services/api/v1/indicators' endpoint that allows to find indicators,
# that endpoint returns a very different object from the one that you can find via the equipment.
# it only has properties 'id', 'dataType', 'description' and 'internalId', which makes it incompatible
# with property mapping above. Therefore, `find_indicators` is not implemented here...



"""Stores and returns OAuth clients."""
import logging

from .OAuthServiceImpl import OAuth2Client

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())

_clients = {}


def get_oauth_client(name) -> OAuth2Client:
    """Return an existing OAuth client or create a new one based on the name."""
    if name in _clients:
        return _clients[name]

    LOG.debug("Creating new OAuth client for '%s'", name)
    _clients[name] = OAuth2Client(name)
    return _clients[name]



"""Module for various utility functions, in particular those related to fetching data from remote oauth endpoints."""

from typing import Union
from collections.abc import Sequence, Iterable
from collections import Counter
from itertools import product
import operator
import logging
import warnings
import re

import pandas as pd
import plotnine as p9

from ..utils.oauth_wrapper import get_oauth_client
from ..utils.plot_helper import _default_plot_theme
from ..utils.config import SailorConfig
from ..utils.utils import DataNotFoundWarning

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())


def _compose_queries(unbreakable_filters, breakable_filters):
    # So the AC endpoints can only accept a certain URL length
    # and since the filters are part of the URL for GET requests
    # we have to make sure the query doesn't get too long.
    # Assume we have a query like (A AND B AND (C OR D) AND (E OR F OR G))
    # (where the 'OR' terms can become quite long).
    # The shortest way to break this up is into a number of sequential query
    # like ((A AND B AND C AND E) OR (A AND B AND C AND F) OR ...) where we can
    # break on each OR.
    # If any of these individual terms is longer than the allowed length we can't
    # execute the query.
    #
    # However, while this way of breaking it gives us the shortest possible URLs
    # it also gives us the largest amount of queries. So it would be better to break
    # it into e.g. ((A AND B AND (C OR D) AND (E OR F)) OR (A AND B AND (C OR D) AND G)
    # which gives us two queries. Unfortunately I can't think of a sane (and non-exponential)
    # algorithms to find this optimal breakup.
    #
    # So my compromise is to include all the (OR) groups that can fit as a whole in the query,
    # and to break the remaining part up into cartesian products.

    if not (unbreakable_filters or breakable_filters):
        return []

    max_filter_length = 2000

    filter_string = ' and '.join(unbreakable_filters)
    current_fixed_length = len(filter_string)

    cartesian_product = set(product(*breakable_filters))
    remaining_cartesian_length_max = max(len(' and '.join(p)) for p in cartesian_product)

    if max_filter_length < current_fixed_length + remaining_cartesian_length_max + len(' and '):
        raise RuntimeError('Your filter conditions are too complex. Please split your query into multiple calls.')

    breakable_filters = sorted(breakable_filters, key=len)

    # add entire groups for as long as we can
    for idx in range(len(breakable_filters)):
        subfilter = '(' + ' or '.join(breakable_filters[idx]) + ')'
        cartesian_product = set(product(*breakable_filters[(idx + 1):]))
        remaining_cartesian_length_max = max(len(' and '.join(p)) for p in cartesian_product)

        if max_filter_length > len(subfilter) + current_fixed_length + remaining_cartesian_length_max + 2*len(' and '):
            filter_string = ' and '.join([filter_string, subfilter])
            if filter_string.startswith(' and '):
                filter_string = filter_string[5:]
            current_fixed_length = len(filter_string)
        else:
            LOG.debug('Can not fit next filter term completely, breaking at %s', idx)
            break
    else:
        # did not break, so we could process all breakable filters here
        return [filter_string]

    # add everything that's left as cartesian product
    # as a small optimisation: everything but the *last* group.
    # we'll split the last group again, to fill up the space
    # we have as much as possible
    if idx < len(breakable_filters) - 1:  # more than one group left
        cartesian_product = set(product(*breakable_filters[idx:-1]))
        filters = [filter_string + ' and ' + ' and '.join(p) for p in cartesian_product]
    else:
        filters = [filter_string]

    # now we add the last group, in chunks
    if idx < len(breakable_filters):
        current_max_length = max(len(query) for query in filters)
        remaining_length = max_filter_length - current_max_length - len(' and ')
        filter_group = breakable_filters[-1]

        result = []
        start_idx = 0
        end_idx = 2
        while end_idx <= len(filter_group):
            subfilter = '(' + ' or '.join(filter_group[start_idx:end_idx]) + ')'

            # this one is too long, we need to break one before
            if len(subfilter) > remaining_length:
                last_subfilter = '(' + ' or '.join(filter_group[start_idx:(end_idx - 1)]) + ')'
                result.extend(q + ' and ' + last_subfilter if q != '' else last_subfilter for q in filters)
                # this one is too long, but also the last element
                # since we had to break, we have to add the last element separately
                if end_idx == len(filter_group):
                    result.extend(q + ' and ' + filter_group[-1] for q in filters)
                    break
                else:
                    start_idx = end_idx - 1

            # we've reached the and, but didn't neet to break. add everything we have left
            if end_idx == len(filter_group):
                result.extend(q + ' and ' + subfilter if q != '' else subfilter for q in filters)
                break

            end_idx += 1

        filters = result

    return filters


def _fetch_data(endpoint_url, unbreakable_filters=(), breakable_filters=()):
    """Retrieve data from the AssetCentral service."""
    filters = _compose_queries(unbreakable_filters, breakable_filters)
    oauth_client = get_oauth_client('asset_central')

    if not filters:
        filters = ['']

    result = []
    for filter_string in filters:
        params = {'$filter': filter_string} if filter_string else {}
        params['$format'] = 'json'
        endpoint_data = oauth_client.request('GET', endpoint_url, params=params)

        if isinstance(endpoint_data, list):
            result.extend(endpoint_data)
        else:
            result.append(endpoint_data)

    if len(result) == 0:
        warnings.warn(DataNotFoundWarning(), stacklevel=2)

    return result


def _add_properties(cls):
    """Add AssetCentral properties to a class based on the property mapping defined in the class."""
    property_map = cls.get_property_mapping()
    for our_name, v in property_map.items():
        their_name, getter, setter, deleter = v

        # the assignment of the default value (`key=their_name`)
        # is necessary due to the closure rules in loops
        def _getter(self, key=their_name):
            return self.raw.get(key, None)
        if getter is None:
            getter = _getter

        setattr(cls, our_name, property(getter, setter, deleter))
    return cls


def _unify_filters(equality_filters, extended_filters, property_mapping):
    operator_map = {
        '>': 'gt',
        '<': 'lt',
        '>=': 'ge',
        '<=': 'le',
        '!=': 'ne',
        '==': 'eq'
    }

    if equality_filters is None:
        equality_filters = {}
    if extended_filters is None:
        extended_filters = []
    if property_mapping is None:
        property_mapping = {}

    unified_filters = []
    not_our_term = []
    for k, v in equality_filters.items():
        if k in property_mapping:
            key = property_mapping[k][0]
        else:
            key = k
            not_our_term.append(key)

        def quote_if_string(x):
            if isinstance(x, str):
                return f"'{x}'"
            else:
                return str(x)

        if _is_non_string_iterable(v):
            v = [quote_if_string(x) for x in v]
        else:
            v = quote_if_string(v)

        unified_filters.append((key, 'eq', v))

    quoted_pattern = re.compile(r'^(\w+) *?(>|<|==|<=|>=|!=) *?([\"\'])((?:\\?.)*?)\3$')
    unquoted_pattern = re.compile(r'^(\w+) *?(>|<|==|<=|>=|!=) *?([+-]?[\w.]+)$')
    for filter_entry in extended_filters:
        if match := quoted_pattern.fullmatch(filter_entry):
            k, o, _, v = match.groups()
            v = f"'{v}'"  # we always need single quotes, but want to accept double quotes as well, hence re-writing.
        elif match := unquoted_pattern.fullmatch(filter_entry):
            k, o, v = match.groups()
            if v in property_mapping:
                v = property_mapping[v][0]
        else:
            raise RuntimeError(f'Failed to parse filter entry {filter_entry}')

        if k in property_mapping:
            key = property_mapping[k][0]
        else:
            key = k
            not_our_term.append(key)

        unified_filters.append((key, operator_map[o], v))

    if len(not_our_term) > 0:
        warnings.warn(f'Following parameters are not in our terminology: {not_our_term}', stacklevel=3)

    return unified_filters


def _parse_filter_parameters(equality_filters=None, extended_filters=(), property_mapping=None):
    """
    Parse equality and extended filters into breakable and unbreakable filters.

    The distinction between breakable and unbreakable filters is necessary because the AssetCentral endpoints are
    generally queried via `GET` requests with filter parameters passed in the header, and there is a limit to the
    header size. This allows breaking a query into multiple queries of shorter length while still retrieving the
    minimal result set the filters requested.
    """
    unified_filters = _unify_filters(equality_filters, extended_filters, property_mapping)

    breakable_filters = []      # always required
    unbreakable_filters = []    # can be broken into sub-filters if length is exceeded

    for key, op, value in unified_filters:
        if _is_non_string_iterable(value):
            breakable_filters.append([f"{key} {op} {elem}" for elem in value])
        else:
            unbreakable_filters.append(f"{key} {op} {value}")

    return unbreakable_filters, breakable_filters


def _apply_filters_post_request(data, equality_filters, extended_filters, property_mapping):
    """Allow filtering of the results returned by an AssetCentral query if the endpoint doesn't implement `filter`."""
    unified_filters = _unify_filters(equality_filters, extended_filters, property_mapping)
    result = []

    def strip_quote_marks(value):
        if value[0] == "'" and value[-1] == "'":
            return value[1:-1]
        return value

    for elem in data:
        for key, op, value in unified_filters:
            if _is_non_string_iterable(value):
                value = [strip_quote_marks(v) for v in value]
                if elem[key] not in value:
                    break
            else:
                value = strip_quote_marks(value)
                if not getattr(operator, op)(elem[key], value):
                    break
        else:
            result.append(elem)

    return result


def _ac_application_url():
    """Return the Assetcentral application URL from the SailorConfig."""
    return SailorConfig.get('asset_central', 'application_url')


class AssetcentralEntity:
    """Common base class for Assetcentral entities."""

    def __init__(self, ac_json: dict):
        """Create a new entity."""
        self.raw = ac_json

    def __repr__(self) -> str:
        """Return a very short string representation."""
        name = getattr(self, 'name', getattr(self, 'short_description', None))
        return f'"{self.__class__.__name__}(name="{name}", id="{self.id}")'

    def __eq__(self, obj):
        """Compare two objects based on instance type and id."""
        return isinstance(obj, self.__class__) and obj.id == self.id

    def __hash__(self):
        """Hash of an asset central object is the hash of it's id."""
        return self.id.__hash__()


class ResultSet(Sequence):
    """Baseclass to be used in all Sets of AssetCentral objects."""

    _element_type = None
    _method_defaults = {}

    def __init__(self, elements, generating_query_params=None):
        """Create a new ResultSet from the passed elements."""
        self.elements = list(set(elements))
        if len(self.elements) != len(elements):
            duplicate_elements = [k for k, v in Counter(elements).items() if v > 1]
            LOG.info(f'Duplicate elements encountered when creating {type(self).__name__}, discarding duplicates. '
                     f'Duplicates of the following elements were discarded: %s', duplicate_elements)

        self.__generating_query_params = generating_query_params

    def __len__(self) -> int:
        """Return the number of objects stored in the ResultSet to implement the `Sequence` interface."""
        return self.elements.__len__()

    def __eq__(self, other):
        """Two ResultSets are equal if all of their elements are equal (order is ignored)."""
        if isinstance(self, other.__class__):
            return set(self.elements) == set(other.elements)
        return False

    def __getitem__(self, arg: Union[int, slice]):
        """Return a subset of the ResultSet to implement the `Sequence` interface."""
        selection = self.elements.__getitem__(arg)
        if isinstance(arg, int):
            return selection
        else:
            return self.__class__(selection)

    def __add__(self, other):
        """Combine two ResultSets as the sum of all elements, required to implement the `Sequence` interface."""
        if not isinstance(other, type(self)):
            raise TypeError('Only ResultSets of the same type can be added.')
        return self.__class__(self.elements + other.elements, 'set-summation')

    def as_df(self, columns=None):
        """Return all information on the objects stored in the ResultSet as a pandas dataframe."""
        columns = self._element_type.get_property_mapping().keys() if columns is None else columns
        return pd.DataFrame({
            prop: [element.__getattribute__(prop) for element in self.elements] for prop in columns
        })

    def filter(self, **kwargs):
        """Select a subset of the ResultSet based on named filter criteria for the attributes of the elements."""
        selection = []

        for element in self.elements:
            for attribute, value in kwargs.items():
                if _is_non_string_iterable(value) and getattr(element, attribute) not in value:
                    break
                elif not _is_non_string_iterable(value) and getattr(element, attribute) != value:
                    break
            else:
                selection.append(element)
        return self.__class__(selection)

    def plot_distribution(self, by=None, fill=None, dropna=False):
        """
        Plot the distribution of elements of a ResultSet based on their properties.

        This effectively creates a histogram with the number of elements per group on the y-axis, and the group
        (given by the `by` parameter) on the x-axis. Additionally, the fill colour of the bar can be used to
        distinguish a second dimension.
        """
        by = self._method_defaults['plot_distribution']['by'] if by is None else by
        display_name = self._element_type.__name__ + 's'

        columns = [by]
        aes = {'x': by}
        if fill is not None:
            columns.append(fill)
            aes['fill'] = fill

        data = self.as_df(columns)
        if dropna:
            data = data.dropna(subset=[by])
            if len(data) == 0:
                raise RuntimeError(f'No {display_name} with non-empty "{by}" found. Can not create plot.')

        data = data.fillna('NA')
        if data.dtypes[by] == 'O':  # strings/objects, we treat these as categorical
            plot_function = p9.geom_bar()
        else:
            plot_function = p9.geom_histogram(color='white', bins=20)  # anything else, treated as continuous

        plot = (
                p9.ggplot(data, p9.aes(**aes)) +
                plot_function +
                _default_plot_theme() +
                p9.ggtitle(f'Number of {display_name} per {by}')
                )
        return plot


def _is_non_string_iterable(obj):
    if issubclass(obj.__class__, str):
        return False
    return isinstance(obj, Iterable)



"""
Failure Mode module can be used to retrieve FailureMode information from AssetCentral.

Classes are provided for individual FailureModes as well as groups of FailureModes (FailureModeSet).
"""


from .utils import _fetch_data, _add_properties, _parse_filter_parameters, ResultSet, \
    AssetcentralEntity, _ac_application_url
from .constants import VIEW_FAILUREMODES


@_add_properties
class FailureMode(AssetcentralEntity):
    """AssetCentral Failure Mode Object."""

    # Properties (in AC terminology) are:
    # Client, ID, SubClass, StatusText, SubClassDescription, Type, EquipmentsCount, ModelsCount, SparepartsCount,
    # LocationsCount, GroupsCount, SystemsCount, ObjectCount, DisplayID, Source, SourceID, Version, Consume,
    # CategoryCode, CategoryDescription, Causes, ShortDescription, LongDescription, Status, LastChange, MyFailureMode,
    # Owner, ObjectID, PatternID, PatternConfidence, MTTFValue, MTTFUnit, MTTFConfidence, MTTRValue, MTTRUnit,
    # MTTRConfidence, MTBFValue, MTBFUnit, MTBFConfidence, PatternName, PatternImage, ImageID, PrimaryExternalID,
    # FailureModesSearchTerms, TypeCode, DetectionMethod

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('ID', None, None, None),
            'name': ('DisplayID', None, None, None),
            'short_description': ('ShortDescription', None, None, None),
            'long_description': ('LongDescription', None, None, None),
            'status_text': ('StatusText', None, None, None),
        }


class FailureModeSet(ResultSet):
    """Class representing a group of FailureModes."""

    _element_type = FailureMode
    _method_defaults = {
        'plot_distribution': {
            'by': 'status_text',
        },
    }


def find_failure_modes(*, extended_filters=(), **kwargs) -> FailureModeSet:
    """Fetch FailureModes from AssetCentral with the applied filters, return an FailureModeSet.

    This method supports the usual filter criteria, i.e.
    - Any named keyword arguments applied as equality filters, i.e. the name of the FailureMode property is checked
    against the value of the keyword argument. If the value of the keyword argument is an iterable (e.g. a list)
    then all objects matching any of the values in the iterable are returned.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.


    Examples
    --------
    Find all FailureModes with name 'MyFailureMode'::

        find_failure_modes(name='MyFailureMode')

    Find all FailureModes which either have the name 'MyFailureMode' or the name 'MyOtherFailureMode'::

        find_failure_modes(name=['MyFailureMode', 'MyOtherFailureMode'])


    If multiple named arguments are provided then *all* conditions have to match.

    Example
    -------
    Find all FailureModes with name 'MyFailureMode' which also have the short description 'Description'::

        find_failure_modes(name='MyFailureMode', short_description='Description')


    The ``extended_filters`` parameter can be used to specify filters that can not be expressed as an equality. Each
    extended_filter needs to be provided as a string, multiple filters can be passed as a list of strings. As above,
    all filter criteria need to match. Extended filters can be freely combined with named arguments. Here, too, all
    filter criteria need to match for a FailureMode to be returned.

    Example
    -------
    Find all FailureModes where equipment count higher or equal to 5::

        find_failure_modes(extended_filters=['equipments_count >= 5'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, FailureMode.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_FAILUREMODES
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return FailureModeSet([FailureMode(obj) for obj in object_list],
                          {'filters': kwargs, 'extended_filters': extended_filters})



"""
Retrieve Notification information from AssetCentral.

Classes are provided for individual Notifications as well as groups of Notifications (NotificationSet).
"""

import pandas as pd
import plotnine as p9

import sailor.assetcentral.equipment
from .constants import VIEW_NOTIFICATIONS
from .utils import _fetch_data, _add_properties, ResultSet, _parse_filter_parameters,\
    AssetcentralEntity, _ac_application_url
from ..utils.timestamps import _string_to_timestamp_parser
from ..utils.plot_helper import _default_plot_theme


@_add_properties
class Notification(AssetcentralEntity):
    """AssetCentral Notification Object."""

    # Properties (in AC terminology) are: notificationId, shortDescription, status, statusDescription, notificationType,
    # notificationTypeDescription, priority, priorityDescription, isInternal, createdBy, creationDateTime,
    # lastChangedBy, lastChangeDateTime, longDescription, startDate, endDate, malfunctionStartDate, malfunctionEndDate,
    # progressStatus, progressStatusDescription, equipmentId, equipmentName, rootEquipmentId, rootEquipmentName,
    # locationId, breakdown, coordinates, source, operatorId, location, assetCoreEquipmentId, operator, internalId,
    # modelId, proposedFailureModeID, proposedFailureModeDisplayID, proposedFailureModeDesc, confirmedFailureModeID,
    # confirmedFailureModeDesc, confirmedFailureModeDisplayIDs, systemProposedFailureModeID,
    # systemProposedFailureModeDesc, systemProposedFailureModeDisplayID, effectID, effectDisplayID, effectDesc,
    # causeID, causeDisplayID, causeDesc, instructionID, instructionTitle

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('notificationId', None, None, None),
            'name': ('internalId', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'long_description': ('longDescription', None, None, None),
            'breakdown': ('breakdown', None, None, None),  # TODO: turn into boolean?

            'confirmed_failure_mode_id': ('confirmedFailureModeID', None, None, None),
            'confirmed_failure_mode_description': ('confirmedFailureModeDesc', None, None, None),
            'confirmed_failure_mode_name': ('confirmedFailureModeDisplayID', None, None, None),
            'end_date': ('endDate', _string_to_timestamp_parser('endDate'), None, None),
            'equipment_id': ('equipmentId', None, None, None),
            'equipment_name': ('equipmentName', None, None, None),
            'location_id': ('locationID', None, None, None),
            'location_name': ('location', None, None, None),
            'malfunction_end_date': ('malfunctionEndDate', _string_to_timestamp_parser('malfunctionEndDate'),
                                     None, None),
            'malfunction_start_date': ('malfunctionStartDate', _string_to_timestamp_parser('malfunctionStartDate'),
                                       None, None),
            'notification_type': ('notificationType', None, None, None),
            'notification_type_description': ('notificationTypeDescription', None, None, None),
            'priority_description': ('priorityDescription', None, None, None),
            'start_date': ('startDate', _string_to_timestamp_parser('startDate'), None, None),
            'status_text': ('statusDescription', None, None, None),
            'system_failure_mode_id': ('systemProposedFailureModeID', None, None, None),
            'system_failure_mode_description': ('systemProposedFailureModeDesc', None, None, None),
            'system_failure_mode_name': ('systemProposedFailureModeDisplayID', None, None, None),
            'user_failure_mode_id': ('proposedFailureModeID', None, None, None),
            'user_failure_mode_description': ('proposedFailureModeDesc', None, None, None),
            'user_failure_mode_name': ('proposedFailureModeDisplayID', None, None, None),
        }

    def plot_context(self, data=None, window_before=pd.Timedelta(days=7), window_after=pd.Timedelta(days=2)):
        """
        Plot a notification in the context of the timeseries data around the time of the notification.

        This plot can be used to gain insight into the sensor behaviour around the time that a malfunction occurs.
        If the `data` parameter is left as `None` the data required for plotting is automatically retrieved from
        SAP IoT.

        Parameters
        ----------
        data
            TimeseriesDataset to use for plotting indicator data near the Notification.
        window_before
            Time interval plotted before a notification. Default value is 7 days before a notification
        window_after
            Time interval plotted after a notification. Default value is 2 days after a notification
        """
        equipment_set = sailor.assetcentral.equipment.find_equipment(id=self.equipment_id)

        if self.start_date and self.end_date:
            data_start = self.start_date - window_before
            area_start = max(data_start, self.start_date)
            data_end = self.end_date + window_after
            area_end = min(data_end, self.end_date)
        elif self.start_date:
            data_start = self.start_date - window_before
            area_start = max(data_start, self.start_date)
            data_end = self.start_date + window_after
            area_end = data_end
        elif self.end_date:
            data_start = self.end_date - window_before
            area_start = data_start
            data_end = self.end_date + window_after
            area_end = min(data_end, self.end_date)
        else:
            raise RuntimeError('Either notification start_date or notification end_date must be known to plot context.')

        if data is None:
            data = equipment_set.get_indicator_data(data_start, data_end)

        plot = (
                data.plot(data_start, data_end, equipment_set=equipment_set) +
                p9.annotate('vline', xintercept=[self.start_date, self.end_date], size=2, linetype='dotted') +
                p9.annotate('rect', xmin=area_start, xmax=area_end,
                            ymin=-float('inf'), ymax=float('inf'), alpha=0.2)
        )
        return plot


class NotificationSet(ResultSet):
    """Class representing a group of Notifications."""

    _element_type = Notification
    _method_defaults = {
        'plot_distribution': {
            'by': 'equipment_name',
        },
    }

    def plot_overview(self):
        """
        Plot an overview over all notifications in the set as a function of time.

        Each notification will be shown by a rectangle, on a y-scale representing the affected equipment
        and with a color representing the confirmed failure mode description.

        Example
        -------
        Plot an overview over all notifications in the dataset "my_notifications" by time::

            my_notifications.plot_overview()
        """
        data = self.as_df(columns=['malfunction_start_date', 'malfunction_end_date',
                                   'equipment_name', 'confirmed_failure_mode_description'])

        aes = {
            'x': 'malfunction_start_date', 'xend': 'malfunction_end_date',
            'y': 'equipment_name', 'yend': 'equipment_name',
            'color': 'confirmed_failure_mode_description',
        }

        plot = p9.ggplot(data, p9.aes(**aes))
        plot += p9.geom_segment(size=6, alpha=0.7)
        plot += _default_plot_theme()

        return plot


def find_notifications(*, extended_filters=(), **kwargs) -> NotificationSet:
    """Fetch Notifications from AssetCentral with the applied filters, return a NotificationSet.

    This method supports the common filter language explained at :ref:`filter`.

    **Allowed entries for filter terms**

        Type of notifications and its meanings
            M1: Maintenance Request, M2: BreakDown

        Priorities and its meanings
            5: Low, 10: Medium, 15: High, 20:  Very High, 25: Emergency

        Status types and its meanings
            NEW: New, PBD: Published, CPT: Completed, IPR: InProcess

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.

    Examples
    --------
    Find all notifications with short_description 'MyNotification'::

        find_notifications(short_description='MyNotification')

    Find all notifications which either have the short_description 'MyNotification' or the short_description
    'MyOtherNotification'::

        find_notifications(short_description=['MyNotification', 'MyOtherNotification'])

    Find all notifications with short_description 'MyNotification' which also have the start date '2020-07-01'::

        find_notifications(short_description='MyNotification', start_date='2020-07-01')

    Find all notifications with a confirmed failure mode description is not empty::

        find_notifications(extended_filters=['confirmed_failure_mode_description != "None"'])

    Find all notifications in a given timeframe for specific equipment::

        find_notifications(extended_filters=['malfunctionStartDate > "2020-08-01"',
                                             'malfunctionEndDate <= "2020-09-01"'],
                           equipment_id=['id1', 'id2'])
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, Notification.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_NOTIFICATIONS
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return NotificationSet([Notification(obj) for obj in object_list],
                           {'filters': kwargs, 'extended_filters': extended_filters})



"""setup.py."""
from setuptools import setup, find_packages

with open('requirements.in', 'r') as req_file:
    install_requires = req_file.readlines()

with open('README.rst', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open('version.txt', 'r', encoding='utf-8') as fh:
    version = fh.read().rstrip()

setup(
    name='sailor',
    version=version,
    author='DSC Data Science Team',
    author_email='project.sailor@sap.com',
    url='https://github.com/SAP/project-sailor',
    description=('Easily access data from your SAP Digital Supply Chain software products for data science projects '
                 'like predictive maintenance or master data analysis.'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    keywords='',
    license='',
    packages=find_packages(include=('sailor', 'sailor.*')),
    python_requires='>=3.8',
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    project_urls={  # Optional
        'Documentation': 'https://sap.github.io/project-sailor',
        'Blog': 'https://blogs.sap.com/',
    },
)



"""
Retrieve Equipment information from AssetCentral.

Classes are provided for individual Equipment as well as groups of Equipment (EquipmentSet).
"""
from __future__ import annotations

from typing import Union, TYPE_CHECKING
from datetime import datetime

import pandas as pd

from sailor import sap_iot
from .constants import VIEW_EQUIPMENT, VIEW_OBJECTS
from .failure_mode import find_failure_modes
from .indicators import Indicator, IndicatorSet
from .notification import find_notifications
from .location import Location, find_locations
from .workorder import find_workorders
from .utils import _fetch_data, _add_properties, _parse_filter_parameters, AssetcentralEntity, ResultSet, \
    _ac_application_url, _apply_filters_post_request
from ..utils.timestamps import _string_to_timestamp_parser

if TYPE_CHECKING:
    from ..sap_iot import TimeseriesDataset
    from .notification import NotificationSet
    from .failure_mode import FailureModeSet
    from .workorder import WorkorderSet


@_add_properties
class Equipment(AssetcentralEntity):
    """AssetCentral Equipment Object."""

    # Properties (in AC terminology) are:
    # equipmentId, name, internalId, status, statusDescription, version, hasInRevision,
    # modelId, modelName, shortDescription, templateId, subclass, modelTemplate,
    # location, criticalityCode, criticalityDescription, manufacturer, completeness, createdOn,
    # changedOn, publishedOn, serialNumber, batchNumber, tagNumber, lifeCycle, lifeCycleDescription,
    # source, imageURL, operator, coordinates, installationDate, equipmentStatus, buildDate,
    # isOperatorValid, modelVersion, soldTo, image, consume, dealer, serviceProvider, primaryExternalId,
    # equipmentSearchTerms, sourceSearchTerms, manufacturerSearchTerms, operatorSearchTerms, class

    _location = None

    @classmethod
    def get_property_mapping(cls):
        """Return a mapping from assetcentral terminology to our terminology."""
        return {
            'id': ('equipmentId', None, None, None),
            'name': ('name', None, None, None),
            'short_description': ('shortDescription', None, None, None),
            'batch_number': ('batchNumber', None, None, None),
            'build_date': ('buildDate', _string_to_timestamp_parser('buildDate', 'ms'), None, None),
            'criticality_description': ('criticalityDescription', None, None, None),
            'model_id': ('modelId', None, None, None),
            'model_name': ('modelName', None, None, None),
            'installation_date': ('installationDate', _string_to_timestamp_parser('installationDate', 'ms'),
                                  None, None),
            'lifecycle_description': ('lifeCycleDescription', None, None, None),
            'location_name': ('location', None, None, None),
            'manufacturer': ('manufacturer', None, None, None),
            'operator': ('operator', None, None, None),
            'serial_number': ('serialNumber', None, None, None),
            'status_text': ('statusDescription', None, None, None),
            'template_id': ('templateId', None, None, None),
        }

    @property
    def location(self) -> Location:
        """Return the Location associated with this Equipment."""
        if self._location is None:
            locations = find_locations(name=self.location_name)  # why do we have a name here, not an ID???
            assert len(locations) == 1
            self._location = locations[0]
        return self._location

    def find_equipment_indicators(self, *, extended_filters=(), **kwargs) -> IndicatorSet:
        """Find all Indicators assigned to this Equipment.

        This method supports the common filter language explained at :ref:`filter`.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all indicators with name 'MyIndicator' for equipment object 'my_equipment'::

            my_equipment.find_equipment_indicators(name='MyIndicator')
        """
        # AC-BUG: this endpoint just silently ignores filter parameters, so we can't really support them...
        endpoint_url = _ac_application_url() + VIEW_EQUIPMENT + f'({self.id})' + '/indicatorvalues'
        object_list = _fetch_data(endpoint_url)

        filtered_objects = _apply_filters_post_request(object_list, kwargs, extended_filters,
                                                       Indicator.get_property_mapping())
        return IndicatorSet([Indicator(obj) for obj in filtered_objects])

    def find_notifications(self, *, extended_filters=(), **kwargs) -> NotificationSet:
        """
        Fetch notifications objects associated with this equipment.

        This is a wrapper for :meth:`sailor.assetcentral.notification.find_notifications`
        that limits the fetch query to this equipment.

        This method supports the common filter language explained at :ref:`filter`.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all notifications for equipment object 'my_equipment'::

            my_equipment.find_notifications()
        """
        kwargs['equipment_id'] = self.id
        return find_notifications(extended_filters=extended_filters, **kwargs)

    def find_failure_modes(self, *, extended_filters=(), **kwargs) -> FailureModeSet:
        """
        Fetch the failure modes configured for the given equipment.

        This method supports the common filter language explained at :ref:`filter`.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Examples
        --------
        Find all failure modes with name 'MyFailureMode' for equipment object 'my_equipment'::

            my_equipment.find_failure_modes(name='MyFailureMode')
        """
        # AC-BUG: this endpoint just silently ignores filter parameters, so we can't really support them...
        if 'id' in kwargs or 'ID' in kwargs:
            raise RuntimeError('Can not manually filter for FailureMode ID when using this method.')
        endpoint_url = _ac_application_url() + VIEW_OBJECTS + 'EQU/' + self.id + '/failuremodes'
        object_list = _fetch_data(endpoint_url)
        kwargs['id'] = [element['ID'] for element in object_list]
        return find_failure_modes(extended_filters=extended_filters, **kwargs)

    def find_workorders(self, *, extended_filters=(), **kwargs) -> WorkorderSet:
        """
        Fetch workorder objects associated with this equipment.

        This is a wrapper for :meth:`sailor.assetcentral.workorder.find_workorders` that limits the fetch query
        to this equipment.

        This method supports the common filter language explained at :ref:`filter`.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Examples
        --------
        Find all workorders for equipment object 'my_equipment'::

            my_equipment.find_workorders()
        """
        kwargs['equipment_id'] = self.id
        return find_workorders(extended_filters=extended_filters, **kwargs)

    def get_indicator_data(self, start: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           end: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           indicator_set: IndicatorSet = None) -> TimeseriesDataset:
        """
        Fetch timeseries data from SAP Internet of Things for Indicators attached to this equipment.

        This is a wrapper for :meth:`sailor.sap_iot.fetch.get_indicator_data` that limits the fetch query
        to this equipment. Note that filtering for the equipment can only be done locally, so calling this function
        repeatedly for different equipment with the same indicators can be very inefficient.

        Parameters
        ----------
        start
            Date of beginning of requested timeseries data. Any time component will be ignored.
        end
            Date of end of requested timeseries data. Any time component will be ignored
        indicator_set
            IndicatorSet for which timeseries data is returned.

        Example
        -------
        Get indicator data for an equipment 'my_equipment' for a period from 01.06.2020 to 05.12.2020 ::
            my_equipment = find_equipment(name='my_equipment')[0]
            my_equipment.get_indicator_data('2020-06-01', '2020-12-05')

        Note
        ----
        If `indicator_set` is not specified, all indicators associated to this equipment are used.
        """
        if start is None or end is None:
            raise ValueError("Time parameters must be specified")

        if indicator_set is None:
            indicator_set = self.find_equipment_indicators()

        return sap_iot.get_indicator_data(start, end, indicator_set, EquipmentSet([self]))


class EquipmentSet(ResultSet):
    """Class representing a group of Equipment."""

    _element_type = Equipment
    _method_defaults = {
        'plot_distribution': {
            'by': 'location_name',
        },
    }

    def find_notifications(self, *, extended_filters=(), **kwargs) -> NotificationSet:
        """Find all Notifications for any of the equipment in this EquipmentSet.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Examples
        --------
        Get all notifications for the 'equipment_set' as a data frame::

            equipment_set = find_equipment()

            equipment_set.find_notifications().as_df()

        Get all Breakdown notifications (M2) for the 'equipment_set' as a data frame::

            equipment_set.find_notifications(type = 'M2').as_df()
        """
        if len(self) == 0:
            raise RuntimeError('This EquipmentSet is empty, can not find notifications.')

        kwargs['equipment_id'] = [equipment.id for equipment in self.elements]
        return find_notifications(extended_filters=extended_filters, **kwargs)

    def find_workorders(self, *, extended_filters=(), **kwargs) -> WorkorderSet:
        """
        Find all Workorders for any of the equipment in this EquipmentSet.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all workorders for an equipment set 'my_equipment_set'::

            my_equipment_set.find_workorders()

        This method supports the common filter language explained at :ref:`filter`.
        """
        if len(self) == 0:
            raise RuntimeError('This EquipmentSet is empty, can not find workorders.')

        kwargs['equipment_id'] = [equipment.id for equipment in self.elements]
        return find_workorders(extended_filters=extended_filters, **kwargs)

    def find_common_indicators(self, *, extended_filters=(), **kwargs) -> IndicatorSet:
        """
        Find all Indicators common to all Equipment in this EquipmentSet.

        This method supports the common filter language explained at :ref:`filter`.

        Parameters
        ----------
        extended_filters
            See :ref:`filter`.
        **kwargs
            See :ref:`filter`.

        Example
        -------
        Find all common indicators for an EquipmentSet 'my_equipment_set'::

            my_equipment_set.find_common_indicators().as_df()


        Note
        ----
        If all the Equipment in the set are derived from the same Model the overlap in
        Indicators is likely very high. If you get fewer indicators than expected from this method
        verify the uniformity of the Equipment included in this set.
        """
        if len(self) == 0:
            raise RuntimeError('This EquipmentSet is empty, can not find common indicators.')

        common_indicators = self.elements[0].find_equipment_indicators(extended_filters=extended_filters, **kwargs)
        for equipment in self.elements[1:]:
            equipment_indicators = equipment.find_equipment_indicators(extended_filters=extended_filters, **kwargs)
            common_indicators = IndicatorSet([indicator for indicator in common_indicators
                                              if indicator in equipment_indicators])

        return common_indicators

    def get_indicator_data(self, start: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           end: Union[str, pd.Timestamp, datetime.timestamp, datetime.date],
                           indicator_set: IndicatorSet = None) -> TimeseriesDataset:
        """
        Fetch timeseries data from SAP Internet of Things for Indicators attached to all equipments in this set.

        This is a wrapper for :meth:`sailor.sap_iot.fetch.get_indicator_data` that limits the fetch query
        to this equipment set.

        Parameters
        ----------
        start
            Date of beginning of requested timeseries data. Any time component will be ignored.
        end
            Date of end of requested timeseries data. Any time component will be ignored.
        indicator_set
            IndicatorSet for which timeseries data is returned.

        Example
        -------
        Get indicator data for all Equipment belonging to the Model 'MyModel'
        for a period from 01.06.2020 to 05.12.2020 ::

            my_equipment_set = find_equipment(model_name='MyModel')
            my_equipment_set.get_indicator_data('2020-06-01', '2020-12-05')
        Note
        ----
        If `indicator_set` is not specified, indicators common to all equipments in this set are used.
        """
        if start is None or end is None:
            raise ValueError("Time parameters must be specified")

        if indicator_set is None:
            indicator_set = self.find_common_indicators()

        return sap_iot.get_indicator_data(start, end, indicator_set, self)


def find_equipment(*, extended_filters=(), **kwargs) -> EquipmentSet:
    """
    Fetch Equipments from AssetCentral with the applied filters, return an EquipmentSet.

    This method supports the common filter language explained at :ref:`filter`.

    Parameters
    ----------
    extended_filters
        See :ref:`filter`.
    **kwargs
        See :ref:`filter`.

    Examples
    --------
    Find all Equipment with the name 'MyEquipment'::

        find_equipment(name='MyEquipment')

    Find all Equipment which either have the name 'MyEquipment' or the name 'MyOtherEquipment'::

        find_equipment(name=['MyEquipment', 'MyOtherEquipment'])

    Find all Equipment with the name 'MyEquipment' which are also located in 'London'::

        find_equipment(name='MyEquipment', location_name='London')

    Find all Equipment installed between January 1, 2018 and January 1, 2019 in 'London'::

        find_equipment(extended_filters=['installationDate >= "2018-01-01"', 'installationDate < "2019-01-01"'],
                        location_name='London')
    """
    unbreakable_filters, breakable_filters = \
        _parse_filter_parameters(kwargs, extended_filters, Equipment.get_property_mapping())

    endpoint_url = _ac_application_url() + VIEW_EQUIPMENT
    object_list = _fetch_data(endpoint_url, unbreakable_filters, breakable_filters)
    return EquipmentSet([Equipment(obj) for obj in object_list],
                        {'filters': kwargs, 'extended_filters': extended_filters})



"""Scope configuration for remote OAuth services."""

#
# Format:
#
# SCOPE_CONFIG = {
#     'service_name_1': ['short_scope_name_1_1', 'short_scope_name_1_2', ...],
#     'service_name_2': ['short_scope_name_2_1', 'short_scope_name_2_2', ...],
#     'service_name_3': ['short_scope_name_3_1', 'short_scope_name_3_2', ...],
#     ...
# }
#
SCOPE_CONFIG = {
    'sap_iot': ['.am.ts.r', '.am.ts.cud', '.r', '.am.map.r', '.export.r']
}
